import json
from typing import Dict, Any, Union


def normalize_headers(action_data: Dict[str, Any]) -> Dict[str, str]:

    normalized_headers = {}
    action_item_headers = action_data.get('headers', {})

    if isinstance(action_item_headers, (str, bytes, bytearray, )):
        action_item_headers = json.loads(action_item_headers)

    for header_name, header in action_item_headers.items():
        normalized_headers[header_name] = header

    return normalized_headers


def parse_data(
    action_data: Dict[str, Any],
    content_type: str
) -> Union[str, Dict[str, Any]]:

    action_item_data = action_data.get('data')
    if isinstance(action_item_data, (str, bytes, bytearray, )) and content_type == 'application/json':
        action_item_data = json.loads(action_item_data)

    return action_item_data


def parse_tags(
    action_data: Dict[str, Any]
):
    action_item_tags = action_data.get('tags', [])

    if isinstance(action_item_tags, (bytes, bytearray, )):
        action_item_tags = action_item_tags.decode()

    if isinstance(action_item_tags, str):
        action_item_tags = action_item_tags.split(',')

    return action_item_tags


from typing import Union
from scipy.stats import burr12
from .base import BaseDistribution


class Burr12Distribution(BaseDistribution):

    def __init__(
        self, 
        size: int,
        c_value: int=2,
        d_value: int=1,
        center: Union[int, float]=0.5,
        randomness: Union[int, float]=0.4
    ):
        super().__init__(
            size=size,
            center=center,
            randomness=randomness,
            frozed_distribution=burr12(
                c_value,
                d_value,
                loc=center,
                scale=randomness,
            )
        )


import asyncio
import functools
import os
import signal
import psutil
import uuid
from typing import (
    List, 
    Dict, 
    Any, 
    Callable,
    Coroutine,
    Union,
    Type
)
from concurrent.futures import ThreadPoolExecutor
from hedra.core.engines.client.config import Config
from hedra.core.engines.types.common.results_set import ResultsSet
from hedra.core.engines.types.graphql.action import GraphQLAction
from hedra.core.engines.types.graphql_http2.action import GraphQLHTTP2Action
from hedra.core.engines.types.grpc.action import GRPCAction
from hedra.core.engines.types.http.action import HTTPAction
from hedra.core.engines.types.http2.action import HTTP2Action
from hedra.core.engines.types.http3.action import HTTP3Action
from hedra.core.engines.types.playwright.command import PlaywrightCommand
from hedra.core.engines.types.task.task import Task
from hedra.core.engines.types.udp.action import UDPAction
from hedra.core.engines.types.websocket.action import WebsocketAction
from hedra.core.hooks.types.action.hook import ActionHook
from hedra.data.connectors.common.connector_type import ConnectorType
from hedra.data.connectors.common.execute_stage_summary_validator import ExecuteStageSummaryValidator
from hedra.data.parsers.parser import Parser
from hedra.logging import HedraLogger
from .cassandra_connector_config import CassandraConnectorConfig
from .cassandra_load_validator import CassandraLoadValidator
from .schema_set import CassandraSchemaSet


def noop():
    pass


Action = Union[
    GraphQLAction,
    GraphQLHTTP2Action,
    GRPCAction,
    HTTPAction,
    HTTP2Action,
    HTTP3Action,
    PlaywrightCommand,
    Task,
    UDPAction,
    WebsocketAction
]


try:
    from cassandra.cqlengine import columns
    from cassandra.cqlengine import connection
    from cassandra.cqlengine.management import sync_table
    from cassandra.query import dict_factory
    from cassandra.cqlengine.query import ModelQuerySet
    from cassandra.cqlengine.models import Model
    from cassandra.cluster import Cluster
    from cassandra.auth import PlainTextAuthProvider
    has_connector = True

except ImportError:
    columns = object
    connection = object
    sync_table = noop
    dict_factory = noop
    ModelQuerySet = object
    Model = object
    Cluster = object
    PlainTextAuthProvider = object
    has_connector = False


def handle_loop_stop(
    signame, 
    executor: ThreadPoolExecutor, 
    loop: asyncio.AbstractEventLoop
): 
    try:
        executor.shutdown(wait=False, cancel_futures=True) 
        loop.stop()
    except Exception:
        pass


class CassandraConnector:
    connector_type = ConnectorType.Cassandra

    def __init__(
        self, 
        config: CassandraConnectorConfig,
        stage: str,
        parser_config: Config,
    ) -> None:
        self.cluster = None
        self.session = None

        self.hosts = config.hosts
        self.port = config.port or 9042

        self.username = config.username
        self.password = config.password
        self.keyspace = config.keyspace
        
        self.table_name = config.table_name
        
        self.replication_strategy = config.replication_strategy
        self.replication = config.replication       
        self.ssl = config.ssl

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.stage = stage
        self.parser_config = parser_config

        self.logger = HedraLogger()
        self.logger.initialize()

        self._executor = ThreadPoolExecutor(max_workers=psutil.cpu_count(logical=False))
        self._loop = asyncio.get_event_loop()

        self.parser = Parser()

        self._primary_column_types = {
            'ascii': columns.Ascii,
            'bigint': columns.BigInt,
            'blob': columns.Blob,
            'boolean': columns.Boolean,
            'date': columns.Date,
            'datetime': columns.DateTime,
            'decimal': columns.Decimal,
            'duration': columns.Duration,
            'uuid': columns.UUID,
            'integer': columns.Integer,
            'ipaddr': columns.Inet,
            'float': columns.Float,
            'map': columns.Map,
            'smallint': columns.SmallInt,
            'text': columns.Text,
            'time': columns.Time,
            'time_uuid': columns.TimeUUID,
            'tinyint': columns.TinyInt,
            'varint': columns.VarInt,
        }

        self._fields: Dict[str, columns.Column] = {}
        self._columns_factory: Dict[
            str,
            Callable[
                [Dict[str, Any]],
                columns.Column
            ]
        ] = {
            'ascii': lambda column_config: columns.Ascii(
                min_length=column_config.get('min_length', 1),
                **column_config,
            ),
            'bigint': lambda column_config: columns.BigInt(**column_config),
            'blob': lambda column_config: columns.Blob(**column_config),
            'boolean': lambda column_config: columns.Boolean(**column_config),
            'date': lambda column_config: columns.Date(**column_config),
            'datetime': lambda column_config: columns.DateTime(**column_config),
            'decimal': lambda column_config: columns.Decimal(**column_config),
            'duration': lambda column_config: columns.Duration(**column_config),
            'uuid': lambda column_config: columns.UUID(
                primary_key=column_config.get('primary_key', True),
                default=uuid.uuid4,
                **column_config
            ),
            'integer': lambda column_config: columns.Integer(**column_config),
            'ipaddr': lambda column_config: columns.Inet(**column_config),
            'list': lambda column_config: columns.List(
                value_type=self._primary_column_types.get(
                    column_config.get('value_type')
                ),
                **column_config
            ),
            'map': lambda column_config: columns.Map(
                key_type=self._primary_column_types.get(
                    column_config.get('value_type')
                ),
                value_type=self._primary_column_types.get(
                    column_config.get('value_type')
                ),
                **column_config
            ),
            'float': lambda column_config: columns.Float(**column_config),
            'set': lambda column_config: columns.Set(
                value_type=self._primary_column_types.get(
                    column_config.get('value_type')
                ),
                **column_config
            ),
            'smallint': lambda column_config: columns.SmallInt(**column_config),
            'text': lambda column_config: columns.Text(
                min_length=column_config.get('min_length', 1),
                **column_config,
            ),
            'time': lambda column_config: columns.Time(**column_config),
            'time_uuid': lambda column_config: columns.TimeUUID(
                primary_key=column_config.get('primary_key', True),
                default=uuid.uuid1,
                **column_config
            ),
            'tinyint':  lambda column_config: columns.TinyInt(**column_config),
            'tuple': lambda column_config: columns.Tuple(
                **column_config
            ),
            'varint':  lambda column_config: columns.VarInt(**column_config),
        }

        self.schemas = CassandraSchemaSet()

    async def connect(self):

        for signame in ('SIGINT', 'SIGTERM', 'SIG_IGN'):
            self._loop.add_signal_handler(
                getattr(signal, signame),
                lambda signame=signame: handle_loop_stop(
                    signame,
                    self._executor,
                    self._loop
                )
            )

        host_port_combinations = ', '.join([
            f'{host}:{self.port}' for host in self.hosts
        ])

        await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Opening amd authorizing connection to Cassandra Cluster at - {host_port_combinations}')
        await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Opening session - {self.session_uuid}')

        auth = None
        if self.username and self.password:
            auth = PlainTextAuthProvider(self.username, self.password)
        

        self.cluster = Cluster(
            self.hosts,
            port=self.port,
            auth_provider=auth,
            ssl_context=self.ssl
        )

        self.session = await self._loop.run_in_executor(
            None,
            self.cluster.connect
        )

        self.session.row_factory = dict_factory

        await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Connected to Cassandra Cluster at - {host_port_combinations}')

        if self.keyspace is None:
            self.keyspace = 'hedra'

        await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Creating Keyspace - {self.keyspace}')

        keyspace_options = f"'class' : '{self.replication_strategy}', 'replication_factor' : {self.replication}"
        keyspace_query = f"CREATE KEYSPACE IF NOT EXISTS {self.keyspace} WITH REPLICATION = " + "{" + keyspace_options  + "};"

        await self._loop.run_in_executor(
            None,
            self.session.execute,
            keyspace_query
        )

        await self._loop.run_in_executor(
            None,
            self.session.set_keyspace,
            self.keyspace
        )
        if os.getenv('CQLENG_ALLOW_SCHEMA_MANAGEMENT') is None:
            os.environ['CQLENG_ALLOW_SCHEMA_MANAGEMENT'] = '1'

        await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                connection.setup,
                self.hosts, 
                self.keyspace, 
                protocol_version=3
            )
        )

        await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Created Keyspace - {self.keyspace}')

    async def create_schemas(self):
        for schema in self.schemas.action_schemas(self.table_name):
            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    sync_table,
                    schema,
                    keyspaces=[self.keyspace]
                )
            )

        for schema in self.schemas.results_schemas(self.table_name):
            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    sync_table,
                    schema,
                    keyspaces=[self.keyspace]
                )
            )

    async def load_execute_stage_summary(
        self,
        options: Dict[str, Any]={}
    ) -> Coroutine[Any, Any, ExecuteStageSummaryValidator]:
        execute_stage_summary = await self.load_data(
            options=options
        )
        
        return ExecuteStageSummaryValidator(**execute_stage_summary)

    async def load_data(
        self,
        options: Dict[str, Any]={}
    ) -> Coroutine[Any, Any, List[Dict[str, Any]]]:
        
        cassandra_load_request = CassandraLoadValidator(**options)

        table: Type[Model] = options.get('table')
        table_name = options.get('table_name')
        
        if table_name is None:
            table_name = self.table_name

        if table is None:
            
            fields = {
                field_name: self._columns_factory.get(
                    field_config.field_type
                )(
                    field_config.options
                ) for field_name, field_config in cassandra_load_request.fields.items()
            }

            table = type(
                table_name,
                (Model, ),
                fields
            )

        if cassandra_load_request.filters:

            data_rows: ModelQuerySet = await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    table.filter,
                    **cassandra_load_request.filters
                )
            )

        else:
            data_rows: ModelQuerySet = await self._loop.run_in_executor(
                self._executor,
                table.all
            )

        if cassandra_load_request.limit:
            data_rows = await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    data_rows.limit,
                    cassandra_load_request.limit
                )
            )

        return [
            row for row in data_rows
        ]
    
    async def store_actions(
        self,
        actions: List[Action],
        options: Dict[str, Any]={}
    ) -> Coroutine[Any, Any, List[Dict[str, Any]]]:
        pass

    async def load_actions(
        self,
        options: Dict[str, Any]={}
    ) -> Coroutine[Any, Any, List[ActionHook]]:
        
        tables = [
            table for table in self.schemas.action_schemas(self.table_name)
        ]

        actions: List[Dict[str, Any]] = []

        table_results: List[List[Dict[str, Any]]] = await asyncio.gather(*[
            asyncio.create_task(
                self.load_data(
                    options={
                        'table': table,
                        'table_name': table_name,
                        **options
                    }
                )
            ) for table_name, table in tables
        ])

        for table_result in table_results:
            actions.extend(table_result)
  

        return await asyncio.gather(*[
            self.parser.parse_action(
                action_data,
                self.stage,
                self.parser_config,
                options
            ) for action_data in actions
        ])
    
    async def load_results(
        self,
        options: Dict[str, Any]={}
    ) -> Coroutine[Any, Any, ResultsSet]:
        
        tables = [
            table for table in self.schemas.action_schemas(self.table_name)
        ]

        results: List[Dict[str, Any]] = []

        table_results: List[List[Dict[str, Any]]] = await asyncio.gather(*[
            asyncio.create_task(
                self.load_data(
                    options={
                        'table': table,
                        'table_name': table_name
                        **options
                    }
                )
            ) for table_name, table in tables
        ])

        for table_result in table_results:
            results.extend(table_result)

        return ResultsSet({
            'stage_results': await asyncio.gather(*[
                self.parser.parse_result(
                    results_data,
                    self.stage,
                    self.parser_config,
                    options
                ) for results_data in results
            ])
        })
    
    async def close(self):

        await self._loop.run_in_executor(
            self._executor,
            self.cluster.shutdown
        )

        self._executor.shutdown(cancel_futures=True)
    


from pydantic import BaseModel
from hedra.reporting.types.common.types import ReporterTypes


class NetdataConfig(BaseModel):
    host: str='localhost'
    port: int=8125
    reporter_type: ReporterTypes=ReporterTypes.Netdata



import uuid
from typing import List
from hedra.logging import HedraLogger
from hedra.reporting.processed_result.types.base_processed_result import BaseProcessedResult
from hedra.reporting.metric import (
    MetricsSet,
    MetricType
)

try:
    from hedra.reporting.types.statsd import StatsD
    from aio_statsd import DogStatsdClient
    from .dogstatsd_config import DogStatsDConfig
    has_connector = True

except Exception:
    from hedra.reporting.types.empty import Empty as StatsD
    DogStatsdClient = None
    DogStatsDConfig = None
    has_connector = False


class DogStatsD(StatsD):

    def __init__(self, config: DogStatsDConfig) -> None:
        super(DogStatsD, self).__init__(config)
    
        self.host = config.host
        self.port = config.port

        self.connection = DogStatsdClient(
            host=self.host,
            port=self.port
        )

        self.types_map = {
            'total': 'increment',
            'succeeded': 'increment',
            'failed': 'increment',
            'actions_per_second': 'histogram',
            'median': 'gauge',
            'mean': 'gauge',
            'variance': 'gauge',
            'stdev': 'gauge',
            'minimum': 'gauge',
            'maximum': 'gauge',
            'quantiles': 'gauge'
        }

        self.stat_type_map = {
            MetricType.COUNT: 'count',
            MetricType.DISTRIBUTION: 'histogram',
            MetricType.RATE: 'gauge',
            MetricType.SAMPLE: 'gauge'
        }

        self._update_map = {
            'count': lambda: NotImplementedError('DogStatsD does not support counts.'),
            'gauge': self.connection.gauge,
            'sets': lambda: NotImplementedError('DogStatsD does not support sets.'),
            'increment': self.connection.increment,
            'histogram': self.connection.histogram,
            'distribution': self.connection.distribution,
            'timer': self.connection.timer
        }

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.logger = HedraLogger()
        self.logger.initialize()

        self.statsd_type = 'StatsD'

    async def submit_events(self, events: List[BaseProcessedResult]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Events to {self.statsd_type}')

        for event in events:
            time_update_function = self._update_map.get('gauge')
            time_update_function(f'{event.name}_time', event.time)
            
            if event.success:
                success_update_function = self._update_map.get('increment')
                success_update_function(f'{event.name}_success', 1)
            
            else:
                failed_update_function = self._update_map.get('increment')
                failed_update_function(f'{event.name}_failed', 1)

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Events to {self.statsd_type}')

    async def submit_custom(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Custom Metrics to {self.statsd_type}')

        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Custom Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            for custom_metric_name, custom_metric in metrics_set.custom_metrics.items():

                metric_type = self.stat_type_map.get(
                    custom_metric.metric_type,
                    'gauge'
                )

                update_function = self._update_map.get(metric_type)
                update_function(
                    f'{metrics_set.name}_{custom_metric_name}',
                    custom_metric.metric_value
                )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Custom Metrics to {self.statsd_type}')


from enum import Enum


class ConnectorType(Enum):
    AWSLambda='aws_lambda'
    BigTable='bigtable'
    Cassandra='cassandra'
    CosmosDB='cosmosdb'
    CSV='csv'
    GCS='gcs'
    HAR='har'
    JSON='json'
    Kafka='kafka'
    MongoDB='mongodb'
    MySQL='mysql'
    Postgres='postgres'
    Redis='redis'
    S3='s3'
    Snowflake='snowflake'
    SQLite='sqlite'
    TimescaleDB='timescaledb'
    XML='xml'


from hedra.data.connectors.common.connector_type import ConnectorType
from pydantic import BaseModel, StrictStr
from typing import List, Optional


class CSVConnectorConfig(BaseModel):
    filepath: StrictStr
    file_mode: StrictStr='r'
    reporter_type: ConnectorType=ConnectorType.CSV
    headers: Optional[List[StrictStr]]


from typing import Optional, Dict, Union, Tuple, Any
from pydantic import (
    BaseModel, 
    StrictStr, 
    StrictInt, 
    StrictFloat, 
    StrictBool, 
    validator
)


class ActionHookValidator(BaseModel):
    names: Tuple[StrictStr, ...]
    weight: StrictInt
    order: StrictInt
    metadata: Optional[Dict[str, Union[StrictStr, StrictInt, StrictFloat]]]
    skip: StrictBool

    class Config:
        arbitrary_types_allowed = True

    @validator('weight', 'order')
    def validate_weight_and_order(cls, val):
        assert val > 0, "Order and weight values must be greater than zero!"
        return val




import asyncio
import itertools
import networkx
import os
import statistics
import time
import threading
import uuid
from typing import (
    Dict, 
    List, 
    Any,
    Union
)
from hedra.core.graphs.stages.base.exceptions.process_killed_error import ProcessKilledError
from hedra.core.graphs.stages.base.stage import Stage
from hedra.core.graphs.stages.types.stage_types import StageTypes
from hedra.core.graphs.transitions.transition_group import TransitionGroup
from hedra.logging import HedraLogger
from hedra.logging.table.table_types import (
    GraphExecutionResults,
    SystemMetricsCollection,
    GraphResults
)
from hedra.monitoring import(
    CPUMonitor,
    MemoryMonitor
)

from hedra.reporting.metric.metrics_set import MetricsSet
from hedra.reporting.metric.stage_metrics_summary import StageMetricsSummary
from hedra.reporting.system.system_metrics_set import SystemMetricsSet
from hedra.reporting.system.system_metrics_set_types import MonitorGroup
from .transitions import TransitionAssembler, local_transitions
from .status import GraphStatus


class Graph:
    status = GraphStatus.IDLE

    def __init__(
        self, 
        graph_name: str,
        stages: List[Stage], 
        config: Dict[str, Any]={},
        cpus: int=None, 
        worker_id: int=None
    ) -> None:
        self.execution_time = 0
        self.core_config = config

        self.graph_name = graph_name
        self.graph_path = config.get('graph_path')
        self.graph_id = str(uuid.uuid4())
        self.graph_skipped_stages = config.get('graph_skipped_stages', [])

        self.status = GraphStatus.INITIALIZING
        self.graph = networkx.DiGraph()
        self.logger = HedraLogger()
        self._thread_id = threading.current_thread().ident
        self._process_id = os.getpid()
        self.metadata_string = f'Graph - {self.graph_name}:{self.graph_id} - thread:{self._thread_id} - process:{self._process_id} - '

        self.logger.initialize()

        self.logger.hedra.sync.debug(f'{self.metadata_string} - Changed status to - {GraphStatus.INITIALIZING.name} - from - {GraphStatus.IDLE.name}')
        self.logger.filesystem.sync['hedra.core'].info(f'{self.metadata_string} - Changed status to - {GraphStatus.INITIALIZING.name} - {GraphStatus.IDLE.name}')

        self.transitions_graph = []
        self._transitions: List[TransitionGroup] = []
        self._results = None

        self.logger.hedra.sync.debug(f'{self.metadata_string} - Found - {len(stages)} - stages')
        self.logger.filesystem.sync['hedra.core'].debug(f'{self.metadata_string} - Found - {len(stages)} - stages')

        self.stage_types = {
            subclass.stage_type: subclass for subclass in Stage.__subclasses__()
        }

        self.instances: Dict[StageTypes, List[Stage]] = {}
        for stage in self.stage_types.values():
            stage_instances = [stage for stage in stage.__subclasses__() if stage.__module__ == self.core_config.get('graph_module')]
            self.instances[stage.stage_type] = stage_instances

        self.stages: Dict[str, Stage] = {stage.__name__: stage for stage in stages}
        
        self.graph.add_nodes_from([
            (
                stage_name, 
                {"stage": stage}
            ) for stage_name, stage in self.stages.items()
        ])

        for stage in stages:
            
            self.logger.hedra.sync.debug(f'{self.metadata_string} - Adding dependencies for stage - {stage.__name__}')
            self.logger.filesystem.sync['hedra.core'].debug(f'{self.metadata_string} - Adding dependencies for stage - {stage.__name__}')

            for dependency in stage.dependencies:
                if self.graph.nodes.get(dependency.__name__):
                    
                    self.logger.hedra.sync.debug(f'{self.metadata_string} - Adding edge from stage - {dependency.__name__} - to stage - {stage.__name__}')
                    self.logger.filesystem.sync['hedra.core'].debug(f'{self.metadata_string} - Adding edge from stage - {dependency.__name__} - to stage - {stage.__name__}.')

                    self.graph.add_edge(dependency.__name__, stage.__name__)

        self.execution_order = [
            generation for generation in networkx.topological_generations(self.graph)
        ]

        self.runner = TransitionAssembler(
            local_transitions,
            graph_name=self.graph_name,
            graph_path=self.graph_path,
            graph_id=self.graph_id,
            graph_skipped_stages=self.graph_skipped_stages,
            cpus=cpus,
            worker_id=worker_id,
            core_config=self.core_config
        )

    def assemble(self):

        self.status = GraphStatus.ASSEMBLING

        self.logger.hedra.sync.debug(f'{self.metadata_string} - Changed status to - {GraphStatus.ASSEMBLING.name} - from - {GraphStatus.INITIALIZING.name}')
        self.logger.filesystem.sync['hedra.core'].info(f'{self.metadata_string} - Changed status to - {GraphStatus.ASSEMBLING.name} - from - {GraphStatus.INITIALIZING.name}')

        # A user will never specify an Idle stage. Instead, we prepend one to 
        # serve as the source node for a graphs, ensuring graphs have a 
        # valid starting point and preventing the user from forgetting things
        # like a Setup stage.

        self.logger.hedra.sync.debug(f'{self.metadata_string} - Prepending {StageTypes.IDLE.name} stage')
        self.logger.filesystem.sync['hedra.core'].debug(f'{self.metadata_string} - Prepending {StageTypes.IDLE.name} stage')

        self._prepend_stage(StageTypes.IDLE)

        # If we haven't specified an Analyze stage for results aggregation,
        # append one.
        if len(self.instances.get(StageTypes.ANALYZE)) < 1:
            self.logger.hedra.sync.debug(f'{self.metadata_string} - Appending {StageTypes.ANALYZE.name} stage')
            self.logger.filesystem.sync['hedra.core'].debug(f'{self.metadata_string} - Appending {StageTypes.ANALYZE.name} stage')

            self._append_stage(StageTypes.ANALYZE)

        # If we havent specified a Submit stage for save aggregated results,
        # append one.
        if len(self.instances.get(StageTypes.SUBMIT)) < 1:
            self.logger.hedra.sync.debug(f'{self.metadata_string} - Appending {StageTypes.SUBMIT.name} stage')
            self.logger.filesystem.sync['hedra.core'].debug(f'{self.metadata_string} - Appending {StageTypes.SUBMIT.name} stage')

            self._append_stage(StageTypes.SUBMIT)

        # Like Idle, a user will never specify a Complete stage. We append
        # one to serve as the sink node, ensuring all Graphs executed can
        # reach a single exit point.
        self.logger.hedra.sync.debug(f'{self.metadata_string} - Appending {StageTypes.COMPLETE.name} stage')
        self.logger.filesystem.sync['hedra.core'].debug(f'{self.metadata_string} - Appending {StageTypes.COMPLETE.name} stage')

        self._append_stage(StageTypes.COMPLETE)

        self.logger.hedra.sync.debug(f'{self.metadata_string} - Generating graph stages and transitions')
        self.logger.filesystem.sync['hedra.core'].debug(f'{self.metadata_string} - Generating graph stages and transitions')

        self.runner.generate_stages(self.stages)
        self._transitions = self.runner.build_transitions_graph(self.execution_order, self.graph)
        self.runner.map_to_setup_stages(self.graph)
        self.runner.apply_config_to_load_hooks(self.graph)

        self.logger.hedra.sync.debug(f'{self.metadata_string} - Assembly complete')
        self.logger.filesystem.sync['hedra.core'].debug(f'{self.metadata_string} - Assembly complete')

    async def run(self) -> GraphResults:

        cpu_monitor = CPUMonitor()
        memory_monitor = MemoryMonitor()

        execution_start = time.monotonic()

        run_task = asyncio.current_task()

        await self.logger.hedra.aio.debug(f'{self.metadata_string} - Changed status to - {GraphStatus.RUNNING.name} - from - {GraphStatus.ASSEMBLING.name}')
        await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Changed status to - {GraphStatus.RUNNING.name} - from - {GraphStatus.ASSEMBLING.name}')
        
        self.status = GraphStatus.RUNNING

        summary_output: GraphExecutionResults = {}
        submit_stage_system_metrics: SystemMetricsCollection = {}
        graph_system_metrics: MonitorGroup = {}

        for transition_group in self._transitions:

            is_idle_transition = False

            transition_stage_types = [transition.edge.source.stage_type for transition in transition_group]
            is_idle_transition = StageTypes.IDLE in transition_stage_types

            transition_group.sort_and_map_transitions()

            current_stages = ', '.join(
                list(set([transition.from_stage.name for transition in transition_group]))
            )

            self.logger.spinner.logger_enabled = True
            if is_idle_transition:
                self.logger.spinner.logger_enabled = False

            async with self.logger.spinner as status_spinner:
                
                if is_idle_transition is False:
                    await self.logger.spinner.append_message(f"Executing stages - {current_stages}")
                
                    for transition in transition_group:
                        await status_spinner.system.debug(f'{self.metadata_string} - Executing stage Transtition - {transition.transition_id} -  from stage - {transition.from_stage.name} - to stage - {transition.to_stage.name}')
                        await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Executing stage Transition - {transition.transition_id} - from stage - {transition.from_stage.name} - to stage - {transition.to_stage.name}')
                        await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Executing stage - {transition.from_stage.name}:{transition.from_stage.stage_id}')

                results = await transition_group.execute_group()
                
                for transition in transition_group:
                    error = transition.edge.exception

                    if isinstance(error, ProcessKilledError):
                        self.status = GraphStatus.CANCELLED
                        return
                    
                    if error:

                        source_stage = transition.edge.source

                        self.status = GraphStatus.FAILED

                        await status_spinner.system.debug(f'{self.metadata_string} - Changed status to - {GraphStatus.FAILED.name} - from - {GraphStatus.RUNNING.name}')
                        await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Changed status to - {GraphStatus.FAILED.name} - from - {GraphStatus.RUNNING.name}')

                        await status_spinner.system.error(f'{self.metadata_string} - Encountered error executing stage - {source_stage.name}:{source_stage.stage_id}')
                        await self.logger.filesystem.aio['hedra.core'].error(f'{self.metadata_string} - Encountered error executing stage - {source_stage.name}:{source_stage.stage_id}')

                        error_transtiton = self.runner.create_error_transition(
                            source_stage,
                            error
                        )

                        await error_transtiton.execute() 
 
                    if transition.edge.source.stage_type == StageTypes.ANALYZE:
                        stage_name = transition.edge.source.name
                        submit_stage_context = transition.edge.source.context

                        analyze_stage_summary_metrics: Dict[
                            str, 
                            Union[
                                str,
                                Dict[str, StageMetricsSummary], 
                                Dict[str, MetricsSet], 
                                SystemMetricsSet
                            ]
                        ] = submit_stage_context.get('analyze_stage_summary_metrics')

                        if analyze_stage_summary_metrics:
                            summary_output[stage_name] = analyze_stage_summary_metrics
                            stage_system_metrics = analyze_stage_summary_metrics.get('system_metrics', {})

                            graph_system_metrics.update(stage_system_metrics.metrics)

                    if transition.edge.source.stage_type == StageTypes.SUBMIT:
                        stage_name = transition.edge.source.name
                        submit_stage_context = transition.edge.source.context

                        submit_stage_system_metrics_set: SystemMetricsSet = submit_stage_context.get('stage_system_metrics')
                        if submit_stage_system_metrics_set:
                            submit_stage_system_metrics[stage_name] = submit_stage_system_metrics_set
                            graph_system_metrics.update(submit_stage_system_metrics_set.metrics)

                if self.status == GraphStatus.FAILED:
                    status_spinner.finalize()
                    await status_spinner.fail('Error')
                    break
                
                for transition in transition_group:
                    await status_spinner.system.debug(f'{self.metadata_string} - Completed stage Transtition - {transition.transition_id} -  from stage - {transition.from_stage.name} - to stage - {transition.to_stage.name}')


                if is_idle_transition is False:
                    completed_transitions_count = len(results)
                    await status_spinner.system.debug(f'{self.metadata_string} - Completed -  {completed_transitions_count} - transitions')
                    await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Completed -  {completed_transitions_count} - transitions')

                    status_spinner.group_finalize()

                    await status_spinner.ok('✔')

                results = None
            
        if self.status == GraphStatus.RUNNING:
            await self.logger.spinner.system.debug(f'{self.metadata_string} - Changed status to - {GraphStatus.COMPLETE.name} - from - {GraphStatus.RUNNING.name}')
            await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Changed status to - {GraphStatus.COMPLETE.name} - from - {GraphStatus.RUNNING.name}')

            self.status = GraphStatus.COMPLETE

        self.execution_time = execution_start - time.monotonic()

        for transition_group in self._transitions:

            group_cpu_metrics: List[List[Union[int, float]]] = []
            group_memory_metrics: List[List[Union[int, float]]]= []

            for transition in transition_group:

                stage_name = transition.edge.source.name

                transition_system_metrics = graph_system_metrics.get(stage_name)
                if transition_system_metrics:
                    stage_cpu_metrics = graph_system_metrics[stage_name].get('cpu')
                    stage_memory_metrics = graph_system_metrics[stage_name].get('memory')

                    for metrics in stage_cpu_metrics.stage_metrics.values():
                        group_cpu_metrics.append(metrics)
                    
                    for metrics in stage_memory_metrics.stage_metrics.values():
                        group_memory_metrics.append(metrics)

                transition.edge.source.context = None
                transition.edge.destination.context = None
                transition.edge.history = None

            if len(group_cpu_metrics) > 0 and len(group_memory_metrics) > 0:
                group_cpu_metrics = [
                    statistics.median(cpu_usage) for cpu_usage in itertools.zip_longest(
                        *group_cpu_metrics,
                        fillvalue=0
                    )
                ]

                group_memory_metrics = [
                    sum(memory_usage) for memory_usage in itertools.zip_longest(
                        *group_memory_metrics,
                        fillvalue=0
                    )
                ]

            cpu_monitor.collected[self.graph_name].extend(group_cpu_metrics)
            memory_monitor.collected[self.graph_name].extend(group_memory_metrics)
            
            transition_group.destination_groups = None
            transition_group.transitions = None
            transition_group.transitions_by_type = None
            transition_group.edges_by_name = None
            transition_group.adjacency_list = None

        pending_tasks = asyncio.all_tasks()

        for task in pending_tasks:
            if task != run_task and task.cancelled() is False:
                task.cancel()

        cpu_monitor.aggregate_worker_stats()
        memory_monitor.aggregate_worker_stats()

        cpu_monitor.stage_metrics[self.graph_name] = cpu_monitor.collected[self.graph_name]
        memory_monitor.stage_metrics[self.graph_name] = memory_monitor.collected[self.graph_name]

        cpu_monitor.visibility_filters[self.graph_name] = True
        memory_monitor.visibility_filters[self.graph_name] = True

        graph_system_metrics = {
            self.graph_name: {
                'cpu': cpu_monitor,
                'memory': memory_monitor
            }
        }

        system_metrics = SystemMetricsSet(
            graph_system_metrics,
            {}
        )

        if self.status == GraphStatus.COMPLETE:
            system_metrics.generate_system_summaries()

        return {
            'metrics': summary_output,
            'submit_stage_system_metrics': submit_stage_system_metrics,
            'graph_system_metrics': system_metrics
        }

    def cleanup(self):
        for executor in self.runner.executors:
            executor.close()

        for transition_group in self._transitions:
            for executor in transition_group._executors:
                executor.close()

    def _append_stage(self, stage_type: StageTypes):

        appended_stage = self.stage_types.get(stage_type)
        last_cut = self.execution_order[-1]

        appended_stage.dependencies = list()

        for stage_name in last_cut:
            stage = self.stages.get(stage_name)

            appended_stage.dependencies.append(stage)

        self.graph.add_node(appended_stage.__name__, stage=appended_stage)

        for stage_name in last_cut:
            self.graph.add_edge(stage_name, appended_stage.__name__)

        self.execution_order = [
            generation for generation in networkx.topological_generations(self.graph)
        ]

        self.stages[appended_stage.__name__] = appended_stage
        self.instances[appended_stage.stage_type].append(appended_stage)

    def _prepend_stage(self, stage_type: StageTypes):
        prepended_stage = self.stage_types.get(stage_type)
        first_cut = self.execution_order[0]

        prepended_stage.dependencies = list()

        for stage_name in first_cut:
            stage = self.stages.get(stage_name)

            stage.dependencies.append(prepended_stage)

        self.graph.add_node(prepended_stage.__name__, stage=prepended_stage)

        for stage_name in first_cut:
            self.graph.add_edge(prepended_stage.__name__, stage_name)

        self.execution_order = [
            generation for generation in networkx.topological_generations(self.graph)
        ]

        self.stages[prepended_stage.__name__] = prepended_stage
        self.instances[prepended_stage.stage_type].append(stage_type)



from types import FunctionType
from typing import Any, Dict
from .plugin_hook import PluginHook


class PluginRegistrar:

    all: Dict[str, PluginHook] = {}
    module_paths: Dict[str, str] = {}

    def __init__(self, hook_type) -> None:
        self.hook_type = hook_type

    def __call__(self, plugin_hook: FunctionType) -> Any:
        self.module_paths[plugin_hook.__name__] = plugin_hook.__module__
        return self.add_hook(self.hook_type)

    def add_hook(self, hook_type: str):
        def wrap_hook():
            def wrapped_method(func):

                hook_name = func.__qualname__
                hook_shortname = func.__name__

                self.all[hook_name] = PluginHook(
                    hook_name,
                    hook_shortname,
                    func,
                    hook_type=hook_type
                )

                return func
            
            return wrapped_method

        return wrap_hook


def makePluginRegistrar():

    return PluginRegistrar


plugin_registrar = makePluginRegistrar()


from hedra.plugins.types.extension import ExtensionPlugin
from typing import Dict, Union


def get_enabled_extensions() -> Dict[str, Union[ExtensionPlugin, None]]:

    enabled_extensions = {}
    
    return enabled_extensions


import os
from hedra.core.graphs.stages import (
    Submit
)

from hedra.reporting.types import (
    HoneycombConfig
)


class SubmitHoneycombResultsStage(Submit):
    config=HoneycombConfig(
        api_key=os.getenv('HONEYCOMB_API_KEY', ''),
        dataset='results'
    )


import asyncio
import time
import uuid
from typing import (
    Any, 
    Coroutine, 
    Dict, 
    TypeVar
)
from hedra.core.engines.types.common.base_engine import BaseEngine
from hedra.core.engines.types.common.timeouts import Timeouts
from hedra.core.engines.types.common.concurrency import Semaphore
from hedra.plugins.types.engine.action import Action
from hedra.plugins.types.engine.result import Result
from hedra.plugins.types.common.types import PluginHooks
from .connection import CustomConnection
from .pool import CustomPool

A = TypeVar('A')
R = TypeVar('R')

class MercuryCustomClient(BaseEngine[A, R]):

    __slots__ = (
        'session_id',
        'timeouts',
        'registered',
        '_hosts',
        'closed',
        'sem',
        'active',
        'waiter',
        'plugin',
        '_on_connect',
        '_on_execute',
        '_on_close',
        'custom_connection',
        'pool'
    )

    def __init__(
        self, 
        plugin: Any,
        concurrency: int=10**3, 
        timeouts: Timeouts = Timeouts(), 
        reset_connections: bool=False
    ) -> None:

        self.session_id = str(uuid.uuid4())
        self.timeouts = timeouts

        self.registered: Dict[str, CustomConnection] = {}
        self._hosts = {}
        self.closed = False

        self.sem = asyncio.Semaphore(value=concurrency)
        
        self.active = 0
        self.waiter: asyncio.Future = None
        self.plugin: Any = plugin

        self._on_connect = self.plugin.hooks.get(PluginHooks.ON_ENGINE_CONNECT)
        self._on_execute = self.plugin.hooks.get(PluginHooks.ON_ENGINE_EXECUTE)
        self._on_close = self.plugin.hooks.get(PluginHooks.ON_ENGINE_CLOSE)

        self.custom_connection: CustomConnection = lambda reset_connection: CustomConnection(
            security_context=self.plugin.security_context,
            reset_connection=reset_connection,
            on_connect=self._on_connect.call,
            on_execute=self._on_execute.call,
            on_close=self._on_close.call
        )

        self.pool = CustomPool(
            self.custom_connection,
            concurrency, 
            reset_connections=reset_connections
        )

        self.pool.create_pool()

    async def set_pool(self, concurrency: int):
        self.sem = asyncio.Semaphore(value=concurrency)
        self.pool = CustomPool(
            self.custom_connection,
            concurrency, 
            reset_connections=self.pool.reset_connections
        )
        self.pool.create_pool()

    def extend_pool(self, increased_capacity: int):
        self.pool.size += increased_capacity
        for _ in range(increased_capacity):
            self.pool.connections.append(
                self.custom_connection(self.pool.reset_connections)
            )
        
        self.sem = Semaphore(self.pool.size)

    def shrink_pool(self, decrease_capacity: int):
        self.pool.size -= decrease_capacity
        self.pool.connections = self.pool.connections[:self.pool.size]
        self.sem = Semaphore(self.pool.size)

    async def prepare(self, action: Action[A]) -> Coroutine[Any, Any, None]:
        try:
            connection: CustomConnection = self.custom_connection(self.pool.reset_connections)

            if action.use_security_context:
                action.security_context = connection.security_context

            await asyncio.wait_for(
                connection.make_connection(action),
                timeout=self.timeouts.connect_timeout
            )

            if action.is_setup is False:
                action.setup()

            self.registered[action.name] = action

        except Exception as e:       
            raise e

    async def execute_prepared_request(self, action: Action[A]) -> Coroutine[Any, Any, Result[R]]:
 
        result: Result[R] = self.plugin.result(action)
        
        result.times['wait_start'] = time.monotonic()
        self.active += 1
 
        async with self.sem:
            connection = self.pool.connections.pop()
            
            try:

                if action.hooks.listen:
                    event = asyncio.Event()
                    action.hooks.channel_events.append(event)
                    await event.wait()

                if action.hooks.before:
                    action: Action[A] = await self.execute_before(action)
                    action.setup()

                result.times['start'] = time.monotonic()

                result = await asyncio.wait_for(
                    connection.execute(action, result),
                    timeout=self.timeouts.total_timeout
                )

                result.times['complete'] = time.monotonic()
       
                self.pool.connections.append(connection)

                if action.hooks.after:
                    result: Result[R] = await self.execute_after(action, result)
                    action.setup()

                if action.hooks.notify:
                    await asyncio.gather(*[
                        asyncio.create_task(
                            channel(result, action.hooks.listeners)
                        ) for channel in action.hooks.channels
                    ])

                    for listener in action.hooks.listeners: 
                        if len(listener.hooks.channel_events) > 0:
                            listener.setup()
                            event = listener.hooks.channel_events.pop()
                            if not event.is_set():
                                event.set()    

            except Exception as e:
                result.times['complete'] = time.monotonic()
                result.error = str(e)

                self.pool.connections.append(
                    self.custom_connection(self.pool.reset_connections)
                )

            self.active -= 1
            if self.waiter and self.active <= self.pool.size:

                try:
                    self.waiter.set_result(None)
                    self.waiter = None

                except asyncio.InvalidStateError:
                    self.waiter = None

            return result


import psutil
from typing import List, Union, Dict
from hedra.core.experiments.mutations.types.base.mutation import Mutation
from .tracing_config import TracingConfig
from .time_parser import TimeParser


class Config:
    
    def __init__(self, **kwargs) -> None:

        for config_option_name, config_option_value in dict(kwargs).items():
            if config_option_value is None:
                del kwargs[config_option_name]
    
        self.total_time_string = kwargs.get('total_time', '1m')
        parsed_time = TimeParser(self.total_time_string)

        self.log_level = kwargs.get('log_level', 'info')
        self.persona_type = kwargs.get('persona_type', 'default')
        self.total_time = parsed_time.time
        self.batch_size = kwargs.get('batch_size', 1000)
        self.batch_interval = kwargs.get('batch_interval')
        self.action_interval = kwargs.get('action_interval', 0)
        self.optimize_iterations = kwargs.get('optimize_iterations', 0)
        self.optimizer_type = kwargs.get('optimizer_type', 'shg')
        self.batch_gradient = kwargs.get('batch_gradient', 0.1)
        self.cpus = kwargs.get('cpus', psutil.cpu_count(logical=False))
        self.no_run_visuals = kwargs.get('no_run_visuals', False)
        self.connect_timeout = kwargs.get('connect_timeout', 15)
        self.request_timeout = kwargs.get('request_timeout', 60)
        self.reset_connections = kwargs.get('reset_connections')
        self.graceful_stop = kwargs.get('graceful_stop', 1)
        self.optimized = False

        if self.request_timeout > self.total_time:
            self.request_timeout = self.total_time

        self.browser_type = kwargs.get('browser_type', 'chromium')
        self.device_type = kwargs.get('device_type')
        self.locale = kwargs.get('locale')
        self.geolocation = kwargs.get('geolocation')
        self.permissions: List[str] = kwargs.get('permissions', [])
        self.color_scheme = kwargs.get('color_scheme')
        self.group_size = kwargs.get('group_size')
        self.playwright_options = kwargs.get('playwright_options', {})
        self.experiment: Dict[str, Union[str, int, List[float]]] = kwargs.get('experiment', {})
        self.tracing: Union[TracingConfig, None] = kwargs.get('tracing')
        self.mutations: Union[List[Mutation], None] = kwargs.get('mutations', [])
        self.actions_filepaths: Union[Dict[str, str], None] = kwargs.get('actions_filepaths')

    def copy(self):

        trace = None
        if self.tracing:
            trace = self.tracing.copy()

        return Config(**{
            'total_time': self.total_time_string,
            'log_level': self.log_level,
            'persona_type': self.persona_type,
            'batch_size': self.batch_size,
            'batch_interval': self.batch_interval,
            'action_interval': self.action_interval,
            'optimize_iterations': self.optimize_iterations,
            'optimizer_type': self.optimizer_type,
            'batch_gradient': self.batch_gradient,
            'cpus': self.cpus,
            'no_run_visuals': self.no_run_visuals,
            'connect_timeout': self.connect_timeout,
            'request_timeout': self.request_timeout,
            'reset_connections': self.reset_connections,
            'graceful_stop': self.graceful_stop,
            'optimized': self.optimized,
            'browser_type': self.browser_type,
            'device_type': self.device_type,
            'locale': self.locale,
            'geolocation': self.geolocation,
            'permissions': self.permissions,
            'color_scheme': self.color_scheme,
            'group_size': self.group_size,
            'playwright_options': self.playwright_options,
            'experiment': self.experiment,
            'trace': trace,
            'mutations': self.mutations,
            'actions_filepaths': self.actions_filepaths
        })



from collections import defaultdict
from hedra.core.hooks.types.context.hook import ContextHook
from hedra.core.hooks.types.base.hook import Hook
from hedra.core.hooks.types.base.event import BaseEvent
from hedra.core.hooks.types.base.event_types import EventType


class ContextEvent(BaseEvent[ContextHook]):

    def __init__(self, target: Hook, source: ContextHook) -> None:

        super(
            ContextEvent,
            self
        ).__init__(
            target,
            source
        )

        self.event_type = EventType.CONTEXT

    def copy(self):
        context_event = ContextEvent(
            self.target.copy(),
            self.source.copy()
        )

        context_event.execution_path = list(self.execution_path)
        context_event.previous_map = list(self.previous_map)
        context_event.next_map = list(self.next_map)
        context_event.next_args = defaultdict(dict)

        return context_event


import json
from typing import Any, Dict, Union
from hedra.core.engines.types.task.result import TaskResult
from .base_processed_result import BaseProcessedResult


class TaskProcessedResult(BaseProcessedResult):

    def __init__(
        self, 
        stage: str, 
        result: TaskResult
    ) -> None:
        super(
            TaskProcessedResult,
            self
        ).__init__(
            stage,
            result
        )

        self.time = result.complete - result.start
        self.timings = {
            'total': self.time,
            'waiting': result.start - result.wait_start,
            'reading': result.complete - result.write_end
        }

        self.data = result.data

    def to_dict(self) -> Dict[str, Union[str, int, float]]:
        return {
            'name': self.name,
            'stage': self.stage,
            'shortname': self.shortname,
            'checks': [check.__name__ for check in self.checks],
            'error': str(self.error),
            'time': self.time,
            'type': self.type,
            'source': self.source,
            'data': self.data,
            **self.timings
        }
    
    def serialize(self) -> str:
        return json.dumps(
            self.to_dict()
        )


from .generated_aws_lambda_results_stage import SubmitAWSLambdaResultsStage
from .generated_aws_timestream_results_stage import SubmitAWSTimestreamResultsStage
from .generated_bigquery_results_stage import SubmitBigQueryResultsStage
from .generated_bigtable_results_stage import SubmitBigTableResultsStage
from .generated_cassandra_results_stage import SubmitCassandraResultsStage
from .generated_cloudwatch_results_stage import SubmitCloudwatchResultsStage
from .generated_cosmosdb_results_stage   import SubmitCosmosDBResultsStage
from .generated_csv_results_stage import SubmitCSVResultsStage
from .generated_datadog_results_stage import SubmitDatadogResultsStage
from .generated_dogstatsd_results_stage import SubmitDogStatsDResultsStage
from .generated_google_cloud_storage_results_stage import SubmitGoogleCloudStorageResultsStage
from .generated_graphite_results_stage import SubmitGraphiteResultsStage
from .generated_honeycomb_results_stage import SubmitHoneycombResultsStage
from .generated_influxdb_results_stage import SubmitInfluxDBResultsStage
from .generated_json_results_stage import SubmitJSONResultsStage
from .generated_kafka_results_stage import SubmitKafkaResultsStage
from .generated_mongodb_results_stage import SubmitMongoDBResultsStage
from .generated_mysql_results_stage import SubmitMySQLResultsStage
from .generated_netdata_results_stage import SubmitNetdataResultsStage
from .generated_newrelic_results_stage import SubmitNewrelicResultsStage
from .generated_postgres_results_stage import SubmitPostgresResultsStage
from .generated_prometheus_results_stage import SubmitPrometheusResultsStage
from .generated_redis_results_stage import SubmitRedisResultsStage
from .generated_s3_results_stage import SubmitS3ResultsStage
from .generated_snowflake_results_stage import SubmitSnowflakeResultsStage
from .generated_sqlite_results_stage import SubmitSQLiteResultsStage
from .generated_statsd_results_stage import SubmitStatsDResultsStage
from .generated_telegraf_results_stage import SubmitTelegrafResultsStage
from .generated_telegraf_statsd_results_stage import SubmitTelegrafStatsDResultsStage
from .generated_timescaledb_results_stage import SubmitTimescaleDBResultsStage


import json
from pydantic import (
    StrictStr,
    StrictInt,
    Json
)
from typing import Dict, Optional, Union, Literal
from hedra.distributed.models.base.message import Message


class HTTPMessage(Message):
    protocol: StrictStr="HTTP/1.1"
    path: Optional[StrictStr]
    method: Optional[
        Literal[
            "GET", 
            "POST",
            "HEAD",
            "OPTIONS", 
            "PUT", 
            "PATCH", 
            "DELETE"
        ]
    ]
    status: Optional[StrictInt]
    status_message: Optional[StrictStr]
    params: Dict[StrictStr, StrictStr]={}
    headers: Dict[StrictStr, StrictStr]={}
    data: Optional[Union[Json, StrictStr]]

    def prepare_response(self):

        message = 'OK'
        if self.error:
            message = self.error

        head_line = f'HTTP/1.1 {self.status} {message}'
            
        encoded_data: str = ''

        if isinstance(self.data, Message):
            encoded_data = json.dumps(self.data.to_data())

            content_length = len(encoded_data)
            headers = f'content-length: {content_length}'

        elif self.data:
            encoded_data = self.data

            content_length = len(encoded_data)
            headers = f'content-length: {content_length}'

        else:
            headers = 'content-length: 0'

        response_headers = self.headers
        if response_headers:
            for key in response_headers:
                headers = f'{headers}\r\n{key}: {response_headers[key]}'

        return f'{head_line}\r\n{headers}\r\n\r\n{encoded_data}'.encode()


from __future__ import annotations
import uuid
import threading
import os
from typing import Any, List, TypeVar, Union, Dict
from hedra.core.personas.streaming.stream_analytics import StreamAnalytics
from hedra.logging import HedraLogger
from hedra.plugins.types.reporter.reporter_config import ReporterConfig
from .experiment.experiments_collection import (
    ExperimentMetricsCollectionSet,
    ExperimentMetricsCollection
)
from .experiment.experiment_metrics_set import ExperimentMetricsSet
from .experiment.experiment_metrics_set_types import (
    VariantSummary,
    MutationSummary
)
from .types import ReporterTypes
from .types import (
    AWSLambda,
    AWSLambdaConfig,
    AWSTimestream,
    AWSTimestreamConfig,
    BigQuery,
    BigQueryConfig,
    BigTable,
    BigTableConfig,
    Cassandra,
    CassandraConfig,
    Cloudwatch,
    CloudwatchConfig,
    CosmosDB,
    CosmosDBConfig,
    CSV,
    CSVConfig,
    Datadog,
    DatadogConfig,
    DogStatsD,
    DogStatsDConfig,
    GoogleCloudStorage,
    GoogleCloudStorageConfig,
    Graphite,
    GraphiteConfig,
    Honeycomb,
    HoneycombConfig,
    InfluxDB,
    InfluxDBConfig,
    JSON,
    JSONConfig,
    Kafka,
    KafkaConfig,
    MongoDB,
    MongoDBConfig,
    MySQL,
    MySQLConfig,
    Netdata,
    NetdataConfig,
    NewRelic,
    NewRelicConfig,
    Postgres,
    PostgresConfig,
    Prometheus,
    PrometheusConfig,
    Redis,
    RedisConfig,
    S3,
    S3Config,
    Snowflake,
    SnowflakeConfig,
    SQLite,
    SQLiteConfig,
    StatsD,
    StatsDConfig,
    Telegraf,
    TelegrafConfig,
    TelegrafStatsD,
    TelegrafStatsDConfig,
    TimescaleDB,
    TimescaleDBConfig,
    XML,
    XMLConfig
)

ReporterType = TypeVar(
    'ReporterType',
    AWSLambdaConfig,
    AWSTimestreamConfig,
    BigQueryConfig, 
    BigTableConfig, 
    CassandraConfig,
    CloudwatchConfig,
    CosmosDBConfig,
    CSVConfig,
    DatadogConfig,
    DogStatsDConfig,
    GoogleCloudStorageConfig,
    GraphiteConfig,
    HoneycombConfig,
    InfluxDBConfig,
    JSONConfig,
    KafkaConfig,
    MongoDBConfig,
    MySQLConfig,
    NetdataConfig,
    NewRelicConfig,
    PostgresConfig,
    PrometheusConfig,
    RedisConfig,
    S3Config,
    SnowflakeConfig,
    SQLiteConfig,
    StatsDConfig,
    TelegrafConfig,
    TelegrafStatsDConfig,
    TimescaleDBConfig,
    XMLConfig
)


class Reporter:
    reporters = {
            ReporterTypes.AWSLambda: lambda config: AWSLambda(config),
            ReporterTypes.AWSTimestream: lambda config: AWSTimestream(config),
            ReporterTypes.BigQuery: lambda config: BigQuery(config),
            ReporterTypes.BigTable: lambda config: BigTable(config),
            ReporterTypes.Cassandra: lambda config: Cassandra(config),
            ReporterTypes.Cloudwatch: lambda config: Cloudwatch(config),
            ReporterTypes.CosmosDB: lambda config: CosmosDB(config),
            ReporterTypes.CSV: lambda config: CSV(config),
            ReporterTypes.Datadog: lambda config: Datadog(config),
            ReporterTypes.DogStatsD: lambda config: DogStatsD(config),
            ReporterTypes.GCS: lambda config: GoogleCloudStorage(config),
            ReporterTypes.Graphite: lambda config: Graphite(config),
            ReporterTypes.Honeycomb: lambda config: Honeycomb(config),
            ReporterTypes.InfluxDB: lambda config: InfluxDB(config),
            ReporterTypes.JSON: lambda config: JSON(config),
            ReporterTypes.Kafka: lambda config: Kafka(config),
            ReporterTypes.MongoDB: lambda config: MongoDB(config),
            ReporterTypes.MySQL: lambda config: MySQL(config),
            ReporterTypes.Netdata: lambda config: Netdata(config),
            ReporterTypes.NewRelic: lambda config: NewRelic(config),
            ReporterTypes.Postgres: lambda config: Postgres(config),
            ReporterTypes.Prometheus: lambda config: Prometheus(config),
            ReporterTypes.Redis: lambda config: Redis(config),
            ReporterTypes.S3: lambda config: S3(config),
            ReporterTypes.Snowflake: lambda config: Snowflake(config),
            ReporterTypes.SQLite: lambda config: SQLite(config),
            ReporterTypes.StatsD: lambda config: StatsD(config),
            ReporterTypes.Telegraf: lambda config: Telegraf(config),
            ReporterTypes.TelegrafStatsD: lambda config: TelegrafStatsD(config),
            ReporterTypes.TimescaleDB: lambda config: TimescaleDB(config),
            ReporterTypes.XML: lambda config: XML(config)
        }

    def __init__(self, reporter_config: Union[ReporterConfig, ReporterType]) -> None:
        self.reporter_id = str(uuid.uuid4())
        self.logger = HedraLogger()
        self.logger.initialize()

        self.graph_name: str=None
        self.graph_id: str=None
        self.stage_name: str=None
        self.stage_id: str=None
        self.metadata_string: str=None
        self.thread_id = threading.current_thread().ident
        self.process_id = os.getpid()

        if reporter_config is None:
            reporter_config = JSONConfig()

        self.reporter_type = reporter_config.reporter_type


        if isinstance(self.reporter_type, ReporterTypes):
            self.reporter_type_name = self.reporter_type.name.capitalize()

        elif isinstance(self.reporter_type, str):
            self.reporter_type_name = self.reporter_type.capitalize()
            
        self.reporter_config = reporter_config
        
        selected_reporter = self.reporters.get(self.reporter_type)
        if selected_reporter is None:
            self.selected_reporter = JSON(reporter_config)

        else:
            self.selected_reporter = selected_reporter(reporter_config)

    async def connect(self):
        self.metadata_string = f'Graph - {self.graph_name}:{self.graph_id} - thread:{self.thread_id} - process:{self.process_id} - Stage: {self.stage_name}:{self.stage_id} - Reporter: {self.reporter_type_name}:{self.reporter_id} - '
        self.selected_reporter.metadata_string = self.metadata_string

        await self.logger.filesystem.aio.create_logfile('hedra.reporting.log')
        self.logger.filesystem.create_filelogger('hedra.reporting.log')

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Connecting')
        await self.selected_reporter.connect()

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Connected')

    async def submit_experiments(self, experiment_metrics_sets: List[ExperimentMetricsSet]):
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting {len(experiment_metrics_sets)} experiments')

        experiment_metrics: List[ExperimentMetricsCollection] = [
            experiment.split_experiments_metrics() for experiment in experiment_metrics_sets
        ]

        experiments: List[Dict[str, str]] = [
            metrics_collection.experiment for metrics_collection in experiment_metrics
        ]

        variants: List[Dict[str, str]] = []
        variants_summaries: List[VariantSummary] = []
        for metrics_collection in experiment_metrics:
            variants.extend(metrics_collection.variants)
            variants_summaries.extend(metrics_collection.variant_summaries)

        mutations: List[MutationSummary] = []
        mutations_summaries: List[MutationSummary] = []
        for metrics_collection in experiment_metrics:
            mutations.extend(metrics_collection.mutations)
            mutations_summaries.extend(metrics_collection.mutation_summaries)

        experiment_metrics_collection_set = ExperimentMetricsCollectionSet(
            experiments_metrics_fields=ExperimentMetricsSet.experiments_fields(),
            variants_metrics_fields=ExperimentMetricsSet.variants_fields(),
            mutations_metrics_fields=ExperimentMetricsSet.mutations_fields(),
            experiments=experiments,
            variants=variants,
            mutations=mutations,
            experiment_summaries=[
                experiment.experiments_summary for experiment in experiment_metrics_sets
            ],
            variant_summaries=variants_summaries,
            mutation_summaries=mutations_summaries
        )
        

        await self.selected_reporter.submit_experiments(experiment_metrics_collection_set)

        await self.selected_reporter.submit_variants(experiment_metrics_collection_set)

        if len(mutations) > 0:
            await self.selected_reporter.submit_mutations(experiment_metrics_collection_set)


        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted {len(experiments)} experiments')

    async def submit_streams(self, stream_metrics: Dict[str, List[StreamAnalytics]]):
        
        streams_count = len(stream_metrics)

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting {streams_count} streams')
        await self.selected_reporter.submit_streams(stream_metrics)

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted {streams_count} streams')

    async def submit_common(self, metrics: List[Any]):
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting {len(metrics)} shared metrics')
        await self.selected_reporter.submit_common(metrics)

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted {len(metrics)} shared metrics')

    async def submit_events(self, events: List[Any]):
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting {len(events)} events')
        await self.selected_reporter.submit_events(events)

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted {len(events)} events')

    async def submit_metrics(self, metrics: List[Any]):
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting {len(metrics)} metrics')
        await self.selected_reporter.submit_metrics(metrics)

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted {len(metrics)} metrics')

    async def submit_custom(self, metrics: List[Any]):
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting {len(metrics)} custom metrics')
        await self.selected_reporter.submit_custom(metrics)

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted {len(metrics)} custom metrics')

    async def submit_errors(self, metrics: List[Any]):
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting {len(metrics)} errors')
        await self.selected_reporter.submit_errors(metrics)

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted {len(metrics)} errors')

    async def submit_system_metrics(self, metrics: List[Any]):
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting {len(metrics)} system metrics sets')
        await self.selected_reporter.submit_session_system_metrics(metrics)
        await self.selected_reporter.submit_stage_system_metrics(metrics)

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted {len(metrics)} system metrics sets')

    async def close(self):
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Closing')
        await self.selected_reporter.close()

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Closed')


import os
from hedra.core.graphs.stages import (
    Submit
)

from hedra.reporting.types import (
    RedisConfig
)


class SubmitRedisResultsStage(Submit):
    config=RedisConfig(
        host='localhost:6379',
        username=os.getenv('REDIS_USERNAME', ''),
        password=os.getenv('REDIS_PASSWORD', ''),
        events_channel='events',
        metrics_channel='metrics',
        channel_type='pipeline',
        secure=True
    )



import dill
import time
import statistics
from collections import defaultdict
from typing import (
    Generic, 
    List, 
    Union, 
    Any, 
    Dict, 
    Optional,
    Type
)
from typing_extensions import TypeVarTuple, Unpack
from hedra.core.engines.client import Client
from hedra.core.engines.client.config import Config
from hedra.core.engines.types.common.types import RequestTypes
from hedra.core.engines.types.common.results_set import ResultsSet
from hedra.core.engines.types.playwright import (
    MercuryPlaywrightClient,
    ContextConfig
)
from hedra.core.engines.types.registry import registered_engines
from hedra.core.hooks.types.condition.decorator import condition
from hedra.core.hooks.types.context.decorator import context
from hedra.core.hooks.types.event.decorator import event
from hedra.core.hooks.types.base.hook_type import HookType
from hedra.core.hooks.types.internal.decorator import Internal
from hedra.core.hooks.types.action.hook import ActionHook
from hedra.core.hooks.types.task.hook import TaskHook
from hedra.core.graphs.stages.base.stage import Stage
from hedra.core.graphs.stages.base.parallel.partition_method import PartitionMethod
from hedra.core.graphs.stages.base.parallel.stage_priority import StagePriority
from hedra.core.graphs.stages.types.stage_types import StageTypes
from hedra.core.personas.streaming.stream_analytics import StreamAnalytics
from hedra.core.personas.persona_registry import (
    get_persona, 
    registered_personas, 
    DefaultPersona
)
from hedra.data.serializers import Serializer
from hedra.logging import logging_manager
from hedra.monitoring import (
    CPUMonitor,
    MemoryMonitor
)
from hedra.plugins.types.plugin_types import PluginType
from hedra.plugins.types.extension.types import ExtensionType
from hedra.plugins.types.extension.extension_plugin import ExtensionPlugin
from hedra.reporting.reporter import ReporterConfig
from hedra.versioning.flags.types.base.active import active_flags
from hedra.versioning.flags.types.base.flag_type import FlagTypes
from hedra.plugins.extensions import get_enabled_extensions
from .parallel import execute_actions


MonitorResults = Dict[str, List[Union[int, float]]]


T = TypeVarTuple('T')


class Execute(Stage, Generic[Unpack[T]]):
    stage_type=StageTypes.EXECUTE
    priority: Optional[str]=None
    retries: int=0

    def __init__(self) -> None:
        super().__init__()
        self.persona = None
        self.client: Client[Unpack[T]] = Client(
            self.graph_name,
            self.graph_id,
            self.name,
            self.stage_id
        )
        
        self.accepted_hook_types = [ 
            HookType.ACTION,
            HookType.CHANNEL, 
            HookType.CHECK,
            HookType.CONDITION,
            HookType.CONTEXT,
            HookType.EVENT,
            HookType.TASK,
            HookType.TRANSFORM
        ]

        self.concurrent_pool_aware_stages = 0
        self.execution_stage_id = 0
        self.optimized = False
        self.execute_setup_stage = None
        self.requires_shutdown = True
        self.allow_parallel = True

        self.source_internal_events = [
            'get_stage_config'
        ]

        self.internal_events = [
            'get_stage_config',
            'get_stage_plugins',
            'check_has_multiple_workers',
            'run_multiple_worker_jobs',
            'aggregate_multiple_worker_results',
            'setup_single_worker_job',
            'run_single_worker_job',
            'complete'
        ]

        self.priority = self.priority
        if self.priority is None:
            self.priority = 'auto'

        self.priority_level: StagePriority = StagePriority.map(
            self.priority
        )

        self.stage_retries = self.retries
        self.serializer = Serializer()

    @Internal()
    async def run(self):
        await self.setup_events()
        self.dispatcher.assemble_execution_graph()
        await self.dispatcher.dispatch_events(self.name)

    @context()
    async def get_stage_config(
        self,
        execute_stage_stream_configs: List[ReporterConfig] = [],
        execute_stage_setup_config: Config=None,
        execute_stage_setup_by: str=None,
        execute_stage_setup_hooks: List[Union[ActionHook , TaskHook]] = []
    ):
        self.context.ignore_serialization_filters = [
            'execute_stage_setup_hooks',
            'setup_stage_ready_stages',
            'execute_stage_results',
            'execute_stage_streamed_analytics',
            'execute_stage_monitors',
            'session_stage_monitors',
            'execute_stage_loaded_actions'
        ]
        persona_type_name = '-'.join([
            segment.capitalize() for segment in execute_stage_setup_config.persona_type.split('-')
        ])

        await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Executing - {execute_stage_setup_config.batch_size} - VUs over {self.workers} threads for {execute_stage_setup_config.total_time_string} using - {persona_type_name} - persona')
        await self.logger.spinner.append_message(f'Stage {self.name} executing - {execute_stage_setup_config.batch_size} - VUs over {self.workers} threads for {execute_stage_setup_config.total_time_string} using - {persona_type_name} - persona')

        main_monitor_name = f'{self.name}.main'

        cpu_monitor = CPUMonitor()
        memory_monitor = MemoryMonitor()

        cpu_monitor.visibility_filters[main_monitor_name] = False
        memory_monitor.visibility_filters[main_monitor_name] = False

        cpu_monitor.stage_type = StageTypes.EXECUTE
        memory_monitor.stage_type = StageTypes.EXECUTE
        cpu_monitor.is_execute_stage = True
        memory_monitor.is_execute_stage = True

        await cpu_monitor.start_background_monitor(main_monitor_name)
        await memory_monitor.start_background_monitor(main_monitor_name)

        return {
            'execute_stage_stream_configs': execute_stage_stream_configs,
            'execute_stage_setup_hooks': execute_stage_setup_hooks,
            'execute_stage_setup_by': execute_stage_setup_by,
            'execute_stage_setup_config': execute_stage_setup_config,
            'execute_stage_monitors': {
                'cpu': cpu_monitor,
                'memory': memory_monitor
            }
        }
    
    @event('get_stage_config')
    async def collect_loaded_actions(self):

        loaded_actions: List[Union[ActionHook, TaskHook]] = []

        ignore_context_keys = set()

        for context_key, value in self.context.items():

            if context_key != 'execute_stage_setup_hooks':

                if isinstance(value, ActionHook):
                    ignore_context_keys.add(context_key)
                    loaded_actions.append(value)

                elif isinstance(value, list):
                    for item in value:
                        if isinstance(item, ActionHook):
                            ignore_context_keys.add(context_key)
                            loaded_actions.append(item)

                elif isinstance(value, dict):
                    for item in value.values():
                        if isinstance(item, ActionHook):
                            ignore_context_keys.add(context_key)
                            loaded_actions.append(item)

        
        self.context.ignore_serialization_filters.extend(
            list(ignore_context_keys)
        )

        return {
            'execute_stage_loaded_actions': loaded_actions
        }

    @event('get_stage_config')
    async def get_stage_plugins(self):

        engine_plugins = self.plugins_by_type.get(PluginType.ENGINE)
        for plugin_name, plugin in engine_plugins.items():
            registered_engines[plugin_name] = plugin
            await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Loaded Engine plugin - {plugin.name}')

        persona_plugins = self.plugins_by_type.get(PluginType.PERSONA)
        for plugin_name, plugin in persona_plugins.items():
            registered_personas[plugin_name] = lambda plugin_config: plugin(plugin_config)
            await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Loaded Persona plugin - {plugin.name}')
        
        execute_stage_extensions: Dict[str, ExtensionPlugin] = {}

        stage_extension_plugins: List[str] = self.plugins_by_type.get(
            PluginType.EXTENSION, {}
        )

        extension_plugins: Dict[str, Type[ExtensionPlugin]] = get_enabled_extensions()
        for plugin_name in stage_extension_plugins:
            plugin = extension_plugins.get(plugin_name)

            if plugin:
                enabled_plugin = plugin()
                execute_stage_extensions[enabled_plugin.name] = enabled_plugin

        source_stage_plugins = defaultdict(list)
        for plugin in self.plugins.values():
            source_stage_plugins[plugin.type].append(plugin.name)

        return {
            'execute_stage_plugins': source_stage_plugins,
            'execute_stage_extensions': execute_stage_extensions
        }
    
    @event('get_stage_config')
    async def get_stage_experiment(
        self,
        execute_stage_setup_config: Config=None,
    ):

        experiment = execute_stage_setup_config.experiment

        if experiment:

            mutations = []

            if execute_stage_setup_config.mutations:
                mutations = [
                    {
                        'mutation_name': mutation.name,
                        'mutation_chance': mutation.chance,
                        'mutation_targets': mutation.targets,
                        'mutation_type': mutation.mutation_type.name.lower()
                    } for mutation in execute_stage_setup_config.mutations
                ]


            return {
                'execute_stage_experiment': {
                    'experiment_name': experiment.get('experiment_name'),
                    'experiment_randomized': experiment.get('random'),
                    'experiment_variant': {
                        'variant_stage': self.name,
                        'variant_weight': experiment.get('weight'),
                        'variant_distribution_intervals': experiment.get('intervals'),
                        'variant_distribution': experiment.get('distribution'),
                        'variant_distribution_type': experiment.get('distribution_type'),
                        'variant_distribution_interval_duration': experiment.get('interval_duration'),
                        'variant_mutations': mutations
                    }
                }
            }

    @condition(
        'collect_loaded_actions',
        'get_stage_plugins',
        'get_stage_experiment'
    )
    async def check_has_multiple_workers(self):
        return {
            'execute_stage_has_multiple_workers': self.total_pool_cpus > 1
        }

    @event('check_has_multiple_workers')
    async def run_multiple_worker_jobs(
        self,
        execute_stage_loaded_actions: List[ActionHook]=[],
        execute_stage_has_multiple_workers: bool = False,
        execute_stage_setup_config: Config=None,
        execute_stage_plugins: Dict[str, List[Any]]={},
        execute_stage_setup_by: str=None,
        execute_stage_stream_configs: List[ReporterConfig] = []
    ):
        loaded_actions: List[str] = []
        for action_hook in execute_stage_loaded_actions:
            loaded_actions.append(
                self.serializer.serialize_action(action_hook)
            ) 

        if execute_stage_has_multiple_workers:
            await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Starting execution for - {self.workers} workers')

            serializable_context = self.context.as_serializable() 
            results_sets = await self.executor.execute_stage_batch(
                execute_actions,
                [
                    dill.dumps({
                        'graph_name': self.graph_name,
                        'graph_path': self.graph_path,
                        'graph_id': self.graph_id,
                        'enable_unstable_features': active_flags[FlagTypes.UNSTABLE_FEATURE],
                        'source_stage_name': self.name,
                        'logfiles_directory': logging_manager.logfiles_directory,
                        'log_level': logging_manager.log_level_name,
                        'source_stage_context': {
                            context_key: context_value for context_key, context_value in serializable_context
                        },
                        'source_stage_loaded_actions': loaded_actions,
                        'source_setup_stage_name': execute_stage_setup_by,
                        'source_stage_id': self.stage_id,
                        'source_stage_plugins': execute_stage_plugins,
                        'source_stage_config': execute_stage_setup_config,
                        'source_stage_stream_configs': execute_stage_stream_configs,
                        'partition_method': PartitionMethod.BATCHES,
                        'workers': self.workers,
                        'worker_id': idx + 1
                    }) for idx in range(self.workers)
                ]
            )

            await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Completed execution for - {self.workers} workers')    


            return {
                'execute_stage_results': results_sets
            }

    @context('run_multiple_worker_jobs')
    async def aggregate_multiple_worker_results(
        self,
        execute_stage_has_multiple_workers: bool = False,
        execute_stage_results: List[Dict[str, Any]]=[],
        execute_stage_experiment: Optional[Dict[str, Any]]=None,
        execute_stage_setup_config: Config=None,
        execute_stage_monitors: Dict[str, Union[CPUMonitor, MemoryMonitor]]={}
    ):
        if execute_stage_has_multiple_workers:
            execute_stage_streamed_analytics: List[StreamAnalytics] = []
            aggregate_results = []
            elapsed_times = []
            stage_contexts = defaultdict(list)

            stage_cpu_monitor: CPUMonitor = execute_stage_monitors.get('cpu')
            stage_memory_monitor: MemoryMonitor = execute_stage_monitors.get('memory')
            
            for result_set in execute_stage_results:

                aggregate_results.extend(result_set.get('results'))
                elapsed_times.append(result_set.get('total_elapsed'))
                worker_id = result_set.get('worker_id')
                
                streamed_analytics = result_set.get('streamed_analytics')
                if streamed_analytics:
                    execute_stage_streamed_analytics.append(streamed_analytics)

                pipeline_context: Dict[str, Any] = result_set.get('context', {})
                for context_key, context_value in pipeline_context.items():
                    stage_contexts[context_key].append(context_value)

                monitors: Dict[str, MonitorResults] = result_set.get('monitoring', {})
                cpu_monitor = monitors.get('cpu', {})
                memory_monitor = monitors.get('memory', {})
                
                for monitor_name, collection_stats in cpu_monitor.items():
                    stage_cpu_monitor.worker_metrics[worker_id][monitor_name] = collection_stats
                    stage_cpu_monitor.collected[monitor_name].extend(collection_stats)

                for monitor_name, collection_stats in memory_monitor.items():
                    stage_memory_monitor.worker_metrics[worker_id][monitor_name] = collection_stats

            for monitor_name, collection_stats in cpu_monitor.items():
                stage_cpu_monitor.visibility_filters[monitor_name] = True
                
            for monitor_name, collection_stats in memory_monitor.items():
                stage_memory_monitor.visibility_filters[monitor_name] = True

            stage_cpu_monitor.aggregate_worker_stats()
            stage_memory_monitor.aggregate_worker_stats()

            main_monitor_name = f'{self.name}.main'

            await stage_cpu_monitor.stop_background_monitor(main_monitor_name)
            await stage_memory_monitor.stop_background_monitor(main_monitor_name)

            stage_cpu_monitor.close()
            stage_memory_monitor.close()

            stage_cpu_monitor.stage_metrics[main_monitor_name] = stage_cpu_monitor.collected[main_monitor_name]
            stage_memory_monitor.stage_metrics[main_monitor_name] = stage_memory_monitor.collected[main_monitor_name]

            total_results = len(aggregate_results)
            total_elapsed = statistics.mean(elapsed_times)

            await self.logger.filesystem.aio['hedra.core'].info( f'{self.metadata_string} - Completed - {total_results} actions at  {round(total_results/total_elapsed)} actions/second over {round(total_elapsed)} seconds')
            await self.logger.spinner.set_default_message(f'Stage - {self.name} completed {total_results} actions at {round(total_results/total_elapsed)} actions/second over {round(total_elapsed)} seconds')

            stage_name = self.name.lower()

            return {
                stage_name: stage_contexts,
                'execute_stage_streamed_analytics': execute_stage_streamed_analytics,
                'execute_stage_results':  ResultsSet({
                    'stage': self.name,
                    'streamed_analytics': execute_stage_streamed_analytics,
                    'stage_batch_size': execute_stage_setup_config.batch_size,
                    'stage_persona_type': execute_stage_setup_config.persona_type,
                    'stage_workers': self.workers,
                    'stage_optimized': self.optimized,
                    'stage_results': aggregate_results,
                    'total_results': total_results,
                    'total_elapsed': total_elapsed,
                    'experiment': execute_stage_experiment
                }),
                'execute_stage_monitors': {
                    self.name: {
                        'cpu': stage_cpu_monitor,
                        'memory': stage_memory_monitor
                    }
                }
            }

    @event('check_has_multiple_workers')
    async def setup_single_worker_job(
        self,
        execute_stage_loaded_actions: List[ActionHook]=[],
        execute_stage_has_multiple_workers: bool = False,
        execute_stage_setup_config: Config=None,
        execute_stage_extensions: Dict[str, ExtensionPlugin]={}
    ):


        self.hooks[HookType.ACTION].extend(execute_stage_loaded_actions)

        if execute_stage_has_multiple_workers is False:

            for extension in execute_stage_extensions.values():

                if extension.extension_type == ExtensionType.GENERATOR:
                    results = await extension.execute(**{
                        'execute_stage_name': self.name,
                        'execute_stage_hooks': self.hooks,
                        'persona_config': execute_stage_setup_config
                    })

                    execute_stage = results.get('execute_stage')

                    if execute_stage:
                        self.hooks = results.get('execute_stage_hooks')

            persona_config = execute_stage_setup_config
            persona = get_persona(persona_config)
            persona.setup(self.hooks, self.metadata_string)

            action_and_task_hooks: List[Union[ActionHook, TaskHook]] = [
                *self.hooks[HookType.ACTION],
                *self.hooks[HookType.TASK]
            ]

            for hook in action_and_task_hooks:
                if hook.action.type == RequestTypes.PLAYWRIGHT and isinstance(hook.session, MercuryPlaywrightClient):

                    await self.logger.filesystem.aio['hedra.core'].info(f'{self.metadata_string} - Setting up Playwright Session')

                    await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Playwright Session - {hook.session.session_id} - Browser Type: {persona_config.browser_type}')
                    await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Playwright Session - {hook.session.session_id} - Device Type: {persona_config.device_type}')
                    await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Playwright Session - {hook.session.session_id} - Locale: {persona_config.locale}')
                    await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Playwright Session - {hook.session.session_id} - geolocation: {persona_config.geolocation}')
                    await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Playwright Session - {hook.session.session_id} - Permissions: {persona_config.permissions}')
                    await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Playwright Session - {hook.session.session_id} - Color Scheme: {persona_config.color_scheme}')


                    await hook.session.setup(
                        config=ContextConfig(
                            browser_type=persona_config.browser_type,
                            device_type=persona_config.device_type,
                            locale=persona_config.locale,
                            geolocation=persona_config.geolocation,
                            permissions=persona_config.permissions,
                            color_scheme=persona_config.color_scheme,
                            options=persona_config.playwright_options
                        )
                    )

            return {
                'execute_stage_persona': persona
            }


    @context('setup_single_worker_job')
    async def run_single_worker_job(
        self,
        execute_stage_has_multiple_workers: bool = False,
        execute_stage_persona: DefaultPersona=None,
        execute_stage_experiment: Optional[Dict[str, Any]]=None,
        execute_stage_setup_config: Config=None,
        execute_stage_monitors: Dict[str, Union[CPUMonitor, MemoryMonitor]]={}
    ):
        if execute_stage_has_multiple_workers is False:

            execution_results = {}

            stage_cpu_monitor: CPUMonitor = execute_stage_monitors.get('cpu')
            stage_memory_monitor: MemoryMonitor = execute_stage_monitors.get('memory')

            main_monitor_name = f'{self.name}.main'

            start = time.monotonic()

            results = await execute_stage_persona.execute()

            elapsed = time.monotonic() - start

            await stage_cpu_monitor.stop_background_monitor(main_monitor_name)
            await stage_memory_monitor.stop_background_monitor(main_monitor_name)

            stage_cpu_monitor.close()
            stage_memory_monitor.close()

            elapsed_idx = int(elapsed)

            stage_cpu_monitor.trim_monitor_samples(
                main_monitor_name,
                elapsed_idx
            )

            stage_memory_monitor.trim_monitor_samples(
                main_monitor_name,
                elapsed_idx
            )

            await self.logger.filesystem.aio['hedra.core'].info(
                f'{self.metadata_string} - Execution complete - Time (including addtional setup) took: {round(elapsed, 2)} seconds'
            )  

            total_results = len(results)
            total_elapsed = execute_stage_persona.total_elapsed

            await stage_cpu_monitor.stop_background_monitor(main_monitor_name)
            await stage_memory_monitor.stop_background_monitor(main_monitor_name)

            stage_cpu_monitor.close()
            stage_memory_monitor.close()

            stage_cpu_monitor.collected.update(
                execute_stage_persona.cpu_monitor.collected
            )

            stage_memory_monitor.collected.update(
                execute_stage_persona.memory_monitor.collected
            )

            stage_cpu_monitor.stage_metrics[main_monitor_name] = stage_cpu_monitor.collected[main_monitor_name]
            stage_memory_monitor.stage_metrics[main_monitor_name] = stage_memory_monitor.collected[main_monitor_name]


            await self.logger.filesystem.aio['hedra.core'].info( f'{self.metadata_string} - Completed - {total_results} actions at  {round(total_results/total_elapsed)} actions/second over {round(total_elapsed)} seconds')
            await self.logger.spinner.set_default_message(f'Stage - {self.name} completed {total_results} actions at {round(total_results/total_elapsed)} actions/second over {round(total_elapsed)} seconds')

            if self.executor:
                await self.executor.shutdown()

            execution_results.update({
                'execute_stage_streamed_analytics': [
                    execute_stage_persona.streamed_analytics
                ],
                'execute_stage_results': ResultsSet({
                    'stage': self.name,
                    'streamed_analytics': [
                        execute_stage_persona.streamed_analytics
                    ],
                    'stage_batch_size': execute_stage_setup_config.batch_size,
                    'stage_persona_type': execute_stage_setup_config.persona_type,
                    'stage_workers': self.workers,
                    'stage_optimized': self.optimized,
                    'stage_results': results,
                    'total_results': total_results,
                    'total_elapsed': total_elapsed,
                    'experiment': execute_stage_experiment
                }),
                'execute_stage_monitors': {
                    self.name: {   
                        'cpu': stage_cpu_monitor,
                        'memory': stage_memory_monitor
                    }
                }
            })
            
            return execution_results

    @event('aggregate_multiple_worker_results', 'setup_single_worker_job')
    async def complete(self):
        return {}


from typing import List, Callable, Awaitable, Any, Tuple
from hedra.core.hooks.types.base.hook_type import HookType
from hedra.core.hooks.types.base.hook import Hook


class ChannelHook(Hook):

    def __init__(
        self, 
        name: str, 
        shortname: str, 
        call: Callable[..., Awaitable[Any]],
        *names: Tuple[str, ...],
        order: int=1,
        skip: bool=False
    ) -> None:
        super().__init__(
            name, 
            shortname, 
            call, 
            skip=skip,
            order=order,
            hook_type=HookType.CHANNEL,
        )

        self.notifiers: List[Any] = []
        self.listeners: List[Any] = []
        self.names = list(set(names))

    async def call(self, **kwargs):

        if self.skip:
            return kwargs

        result = await super().call(**{
            name: value for name, value in kwargs.items() if name in self.params
        })

        if isinstance(result, dict):
            return {
                **kwargs,
                **result
            }

        return {
            **kwargs,
            self.shortname: result
        }

    def copy(self):
        channel_hook = ChannelHook(
            self.name,
            self.shortname,
            self._call,
            *self.names,
            order=self.order,
            skip=self.skip,
        )

        channel_hook.stage = self.stage

        return channel_hook


from hedra.core.engines.types.common.timeouts import Timeouts
from hedra.core.engines.types.playwright.command import (
    PlaywrightCommand,
    Page,
    Input,
    Options,
    URL
)
from hedra.core.engines.types.playwright.client import (
    MercuryPlaywrightClient, 
    ContextConfig
)
from hedra.core.engines.types.playwright.result import PlaywrightResult
from hedra.core.engines.types.common.types import RequestTypes
from hedra.data.serializers.serializer_types.common.base_serializer import BaseSerializer
from typing import List, Dict, Union, Any


class PlaywrightSerializer(BaseSerializer):

    def __init__(self) -> None:
        super().__init__()

    def action_to_serializable(
        self,
        action: PlaywrightCommand
    ) -> Dict[str, Union[str, List[str]]]:
        
        serialized_action = super().action_to_serializable(action)
        return {
            **serialized_action,
            'type': RequestTypes.PLAYWRIGHT,
            'command': action.command,
            'command_args': action.command_args,
            'url': {
                'location': action.url.location,
                'headers': action.url.headers,
            },
            'page': {
                'selector': action.page.selector,
                'attribute': action.page.attribute,
                'x_coordinate': action.page.x_coordinate,
                'y_coordinate': action.page.y_coordinate,
                'frame': action.page.frame
            },
            'input': {
                'key': action.input.key,
                'text': action.input.text,
                'expression': action.input.expression,
                'args': action.input.args,
                'filepath': action.input.filepath,
                'file': action.input.file,
                'path': action.input.path,
                'option': action.input.option,
                'by_label': action.input.by_label,
                'by_value': action.input.by_value
            },
            'options': {
                'event': action.options.event,
                'option': action.options.option,
                'is_checked': action.options.is_checked,
                'timeout': action.options.timeout,
                'extra': action.options.extra,
                'switch_by': action.options.switch_by
            }
        }
    
    def deserialize_action(
        self,
        action: Dict[str, Any]
    ) -> PlaywrightCommand:
        
        url_config = action.get('url', {})
        metadata = action.get('metadata', {})
        page_config = action.get('page', {})
        input_config = action.get('input', {})
        url_config = action.get('url', {})
        options_config=action.get('options', {})
        
        playwright_command = PlaywrightCommand(
            name=action.get('name'),
            command=action.get('command'),
            page=Page(
                selector=page_config.get('selector'),
                attribute=page_config.get('attribute'),
                x_coordinate=page_config.get('x_coordinate'),
                y_coordinate=page_config.get('y_coordinate'),
                frame=page_config.get('frame', 0)
            ),
            url=URL(
                location=url_config.get('locations'),
                headers=url_config.get('headers', {})
            ),
            input=Input(
                key=input_config.get('key'),
                text=input_config.get('text'),
                expression=input_config.get('expression'),
                args=input_config.get('args'),
                filepath=input_config.get('filepath'),
                file=input_config.get('file'),
                path=input_config.get('path'),
                option=input_config.get('option'),
                by_label=input_config.get('by_label', False),
                by_value=input_config.get('by_value', False)
            ),
            options=Options(
                event=options_config.get('event'),
                option=options_config.get('option'),
                is_checked=options_config.get('is_checked', False),
                timeout=options_config.get('timeout', 10),
                extra=options_config.get('extra', {}),
                switch_by=options_config.get('switch_by', 'url')
            ),
            user=metadata.get('user'),
            tags=metadata.get('tags', [])
        )

        return playwright_command
    
    def deserialize_client_config(self, client_config: Dict[str, Any]) -> MercuryPlaywrightClient:
        playwright_client = MercuryPlaywrightClient(
            concurrency=client_config.get('concurrency'),
            group_size=client_config.get('group_size'),
            timeouts=Timeouts(
                **client_config.get('timeouts', {})
            )
        )

        playwright_client.config = ContextConfig(
            **client_config.get('context_config')
        )
        
        return playwright_client
    
    def result_to_serializable(
        self,
        result: PlaywrightResult
    ) -> Dict[str, Any]:
        
        serialized_result = super().result_to_serializable(result)

        encoded_headers = dict(result.headers)

        return {
            **serialized_result,
            'url': result.url,
            'command': result.command,
            'selector': result.selector,
            'attribute': result.attribute,
            'x_coord': result.x_coord,
            'y_coord': result.y_coord,
            'frame': result.frame,
            'type': result.type,
            'headers': encoded_headers,
            'tags': result.tags,
            'user': result.user,
            'key': result.key,
            'text': result.text,
            'expression': result.expression,
            'args': result.args,
            'filepath': result.filepath,
            'file': result.file,
            'option': result.option,
            'event': result.event,
            'timeout': result.option,
            'is_checked': result.is_checked,
            'error': str(result.error)
        }
    
    def deserialize_result(
        self,
        result: Dict[str, Any]
    ) -> PlaywrightResult:
        
        playwright_command = PlaywrightCommand(
            result.get('name'),
            result.get('command'),
            page=Page(
                selector=result.get('selector'),
                x_coordinate=result.get('x_coord'),
                y_coordinate=result.get('y_coord'),
                frame=result.get('frame')
            ),
            url=URL(
                location=result.get('url'),
                headers=result.get('headers')
            ),
            input=Input(
                key=result.get('key'),
                text=result.get('text'),
                expression=result.get('expression'),
                args=result.get('args'),
                filepath=result.get('filepath'),
                file=result.get('file'),
                path=result.get('path'),
                option=result.get('option'),
                by_label=result.get('by_label'),
                by_value=result.get('by_value')
            ),
            options=Options(
                event=result.get('event'),
                is_checked=result.get('is_checked'),
                timeout=result.get('timeout')
            ),
            user=result.get('user'),
            tags=result.get('tags')
        )

     
        playwright_result = PlaywrightResult(
            playwright_command,
            error=result.get('error')
        )

        playwright_result.checks = result.get('checks')
        playwright_result.wait_start = result.get('wait_start')
        playwright_result.start = result.get('start')
        playwright_result.connect_end = result.get('connect_end')
        playwright_result.write_end = result.get('write_end')
        playwright_result.complete = result.get('complete')
        
        playwright_result.type= RequestTypes.PLAYWRIGHT

        return playwright_result
    


from __future__ import annotations
import asyncio
import networkx
from collections import OrderedDict, defaultdict
from typing import List, Union, Dict, Any, Tuple
from hedra.core.hooks.types.action.event import ActionEvent
from hedra.core.hooks.types.task.event import TaskEvent
from hedra.core.hooks.types.channel.event import ChannelEvent
from .event import BaseEvent
from .event_types import EventType
from .simple_context import SimpleContext


class EventDispatcher:

    def __init__(self, timeout: Union[int, float]=None) -> None:
        self.events: OrderedDict[EventType, List[BaseEvent]]= OrderedDict()
        self.priority_map = {
            EventType.CONTEXT: 0,
            EventType.LOAD: 1,
            EventType.EVENT: 2,
            EventType.TRANSFORM: 2,
            EventType.CONDITION: 2,
            EventType.SAVE: 3,
            EventType.ACTION: 4,
            EventType.TASK: 4,
            EventType.CHECK: 5,
            EventType.CHANNEL: 5,
            EventType.METRIC: 6
        }

        event_orderings = list(sorted(
            list(self.priority_map.items()),
            key=lambda event: event[1]
        ))

        for event_type_name, _ in event_orderings:
            self.events[event_type_name] = []
            
        self.events_by_name: Dict[str, BaseEvent] = {}
        self.timeout = timeout
        self.initial_events: Dict[str, List[BaseEvent]]= defaultdict(list)
        self.actions_and_tasks: Dict[str, Union[ActionEvent, TaskEvent]] = {}
        self.channels: Dict[str, ChannelEvent] = {}
        self.skip_list = []
        self.source_name = None
        self.graph: networkx.DiGraph = networkx.DiGraph()

    def __iter__(self):
        for event_type in self.events:
            for event in self.events[event_type]:
                yield event

    def __getitem__(self, event_type: EventType):
        return self.events[event_type]
    
    def copy(self):
        dispatcher = EventDispatcher(timeout=self.timeout)

        for event in self.events_by_name.values():
            dispatcher.add_event(
                event.copy()
            )

        for event in self.events_by_name.values():
            event.context = SimpleContext()
            event.source.context = SimpleContext()

            for graph_event_name in event.events:

                if graph_event_name in dispatcher.events_by_name:
                    dispatcher.events_by_name[event.event_name].events[graph_event_name] = dispatcher.events_by_name.get(graph_event_name)

                else:
                    dispatcher.events_by_name[event.event_name].events[graph_event_name] = event.events.get(graph_event_name)

        for stage in self.initial_events:
            initial_event_names = [
                event.event_name for event in self.initial_events[stage]
            ]

            dispatcher.initial_events[stage] = [
                event for event in dispatcher.events_by_name.values() if event.event_name in initial_event_names
            ]

        dispatcher.assemble_action_and_task_subgraphs()
        dispatcher.assemble_execution_graph()

        return dispatcher

    def set_events(self, events: List[BaseEvent]) -> None:
        for event in events:
            self.events_by_name[event.event_name] = event
            self.events[event.event_type].append(event)

    def add_event(self, event: BaseEvent):

        if isinstance(event, (ActionEvent, TaskEvent)):

            self.actions_and_tasks[event.event_name] = event

        elif isinstance(event, ChannelEvent):
            self.channels[event.event_name] = event
            self.skip_list.append(event.event_name)

        self.events_by_name[event.event_name] = event
        self.events[event.event_type].append(event)

        self.graph.add_node(
            event.event_name,
            event=event
        )

    def assemble_execution_graph(self):
        for event in self.events_by_name.values():
            for target in event.next_map:
                self.graph.add_edge(event.event_name, target)

    def assemble_action_and_task_subgraphs(self):
        for event_name, event in self.actions_and_tasks.items():
            self.skip_list.append(event_name)

            for dependency_name in event.source.names:
                self.skip_list.append(dependency_name)
                
                dependency_event = self.events_by_name.get(dependency_name)
                # If we specify a Channel hook as a dependency of an Action
                # or Task - that Action/Task listens to the Channel.
                if isinstance(dependency_event, ChannelEvent):
                    event.source.is_listener = True
                    dependency_event.source.listeners.append(event)

                event.source.before.extend(
                    self._prepend_action_or_task_event(
                        dependency_event,
                        event
                    )
                )

        for event_name, event in self.events_by_name.items():
            for dependency_name in event.source.names:
                action_or_task_event = self.actions_and_tasks.get(dependency_name)
                if action_or_task_event:
                    self.skip_list.append(event_name)
                    action_or_task_event.source.after.extend(
                        self._append_action_or_task_event(
                            event,
                            action_or_task_event
                        )
                    )
                
                if isinstance(event, ChannelEvent):
                    action_or_task_event.source.is_notifier = True
                    action_or_task_event.source.channels.append(dependency_event)
                    event.source.notifiers.append(action_or_task_event)
                
                    
        for event in self.actions_and_tasks.values():
            for idx, layer in enumerate(event.before):
                event.source.before[idx] = [
                    self.events_by_name.get(event_name) for event_name in layer
                ]

            for idx, layer in enumerate(event.after):
                event.source.after[idx] = [
                    self.events_by_name.get(event_name) for event_name in layer
                ]

            for idx, layer in enumerate(event.checks):
                event.checks[idx] = [
                    self.events_by_name.get(event_name) for event_name in layer
                ]

        # Add listeners for a channel to the channel's notifiers
        for channel_event_name, channel_event in self.channels.items():
            for notifier in channel_event.source.notifiers:
                notifier.listeners.extend(channel_event.listeners)

            self.channels[channel_event_name] = channel_event

    async def dispatch_events(self, stage_name: str):

        await self._execute_batch()       

    def _prepend_action_or_task_event(self, dependency_event: BaseEvent, action_or_task: BaseEvent):
        execution_path = list(dependency_event.execution_path)

        event_layer_found = False
        for idx, layer in enumerate(execution_path):

            if event_layer_found:
                dependency_event.execution_path.remove(layer)

            if action_or_task.event_name in layer:
                event_layer_found = True
                dependency_event.execution_path[idx].remove(action_or_task.event_name)

                if len(layer) < 1:
                    dependency_event.execution_path.remove(layer)

        return dependency_event.execution_path

    def _append_action_or_task_event(self, dependant_event: BaseEvent, action_or_task: BaseEvent):
        execution_path = []

        if dependant_event.event_type == EventType.CHECK:
            for idx, layer in enumerate(dependant_event.execution_path):
                if idx >= len(action_or_task.checks):
                    action_or_task.checks.append(layer)

                else:
                    action_or_task.checks[idx].extend([
                        node for node in layer if node not in action_or_task.checks[idx]
                    ])

        else:
            for idx, layer in enumerate(dependant_event.execution_path):
                if idx >= len(action_or_task.after):
                    action_or_task.after.append(layer)

                else:
                    action_or_task.after[idx].extend([
                        node for node in layer if node not in action_or_task.after[idx]
                    ])

        return execution_path

    async def _execute_batch(self):
        for layer in networkx.topological_generations(self.graph):
            layer_events = [
                self.events_by_name.get(event_name) for event_name in layer if event_name not in self.skip_list
            ]

            results: List[Dict[str, Any]] = await asyncio.gather(*[
                asyncio.create_task(
                    asyncio.wait_for(
                        event.call(**event.next_args),
                        timeout=self.timeout
                    ) if self.timeout else event.call()
                ) for event in layer_events
            ])

            result_events: List[Tuple[BaseEvent, Any]] = []
            for result in results:
                for event_name, result in result.items():
                    event = self.events_by_name.get(event_name)
                    result_events.append((event, result))

            for event, result in result_events:
                next_events = [
                    event.events.get(event_name) for event_name in  event.next_map if event.events.get(event_name) is not None
                ]

                event.context.update(event.next_args[event.event_name])
                event.source.stage_instance.context.update(event.next_args[event.event_name])
            

                for next_event in next_events:
                    next_event.context.update(event.context)

                    next_event.next_args[next_event.event_name].update(result)
                    event.context.update(next_event.next_args[next_event.event_name])
                    event.source.stage_instance.context.update(next_event.next_args[next_event.event_name])

            



from hedra.core.engines.types.common.protocols.ping.ping_type import PingTypesMap
from pydantic import (
    BaseModel, 
    StrictStr,
    StrictInt,
    validator
)


class InjectPingValidator(BaseModel):
    ping_type: StrictStr
    timeout: StrictInt

    @validator('ping_type')
    def validate_ping_type(cls, val):
        types_map = PingTypesMap()
        assert types_map.get(val) is not None
        
        return val




from typing import Optional
from pydantic import BaseModel
from hedra.reporting.types.common.types import ReporterTypes

class SnowflakeConfig(BaseModel):
    username: str
    password: str
    organization_id: str
    account_id: str
    private_key: Optional[str]
    warehouse: str
    database: str
    database_schema: str='PUBLIC'
    events_table: str='events'
    metrics_table: str='metrics'
    experiments_table: str='experiments'
    streams_table: str='streams'
    system_metrics_table: str='system_metrics'
    connect_timeout: int=30
    reporter_type: ReporterTypes=ReporterTypes.Snowflake

    class Config:
        arbitrary_types_allowed = True


import asyncio
import networkx
import threading
import os
from typing import (
    List, 
    Dict, 
    Union, 
    Any, 
    Tuple, 
    Coroutine
)
from collections import defaultdict
from hedra.core.hooks.types.base.event_graph import EventGraph
from hedra.core.graphs.stages.base.import_tools import set_stage_hooks
from hedra.core.graphs.stages.base.stage import Stage
from hedra.core.graphs.stages.error import Error
from hedra.core.graphs.stages.types.stage_types import StageTypes
from hedra.core.hooks.types.base.registrar import registrar
from hedra.core.hooks.types.base.hook import Hook, HookType
from hedra.core.hooks.types.load.hook import LoadHook
from hedra.core.graphs.stages.base.parallel.batch_executor import BatchExecutor
from hedra.core.hooks.types.base.simple_context import SimpleContext
from hedra.core.graphs.transitions.exceptions.exceptions import (
    IsolatedStageError, 
    InvalidTransitionError
)
from hedra.logging import HedraLogger
from hedra.plugins.types.engine.engine_plugin import EnginePlugin
from hedra.plugins.types.reporter.reporter_plugin import ReporterPlugin
from hedra.plugins.types.plugin_types import PluginType
from .transition_group import TransitionGroup
from .transition import Transition
from .common.transtition_metadata import TransitionMetadata
from .common.base_edge import BaseEdge


class TransitionAssembler:

    def __init__(
        self, 
        transition_types, 
        graph_name: str=None,
        graph_path: str=None,
        graph_id: str=None,
        graph_skipped_stages: List[str]=[],
        cpus: int=None, 
        worker_id: int=None,
        core_config: Dict[str, Any]={}
    ) -> None:
        self.transition_types: Dict[Tuple[StageTypes, StageTypes], Coroutine] = transition_types
        self.core_config = core_config
        self.graph_name = graph_name
        self.graph_path = graph_path
        self.graph_id = graph_id
        self.graph_skipped_stages: List[str] = graph_skipped_stages
        self.generated_stages = {}
        self.transitions = {}
        self.instances_by_type: Dict[str, List[Stage]] = {}
        self.cpus = cpus
        self.worker_id = worker_id
        self.loop = asyncio.get_event_loop()
        self.hooks_by_type: Dict[HookType, Dict[str,Hook]] = defaultdict(dict)

        self.logging = HedraLogger()
        self.logging.initialize()

        self._thread_id = threading.current_thread().ident
        self._process_id = os.getpid()
        self.all_hooks = []
        self.edges_by_name: Dict[Tuple[str, str], BaseEdge] = {}
        self.adjacency_list: Dict[str, List[Transition]] = defaultdict(list)
        self.execute_stages: List[Stage] = []
        self.streaming_submit_stages: List[Stage] = []
        self.executors: List[BatchExecutor] = []
        self.all_paths: Dict[str, List[str]] = {}

        self._graph_metadata_log_string = f'Graph - {self.graph_name}:{self.graph_id} - thread:{self._thread_id} - process:{self._process_id} - '

    def generate_stages(self, stages: Dict[str, Stage]) -> None:
    
        stages_count = len(stages)
        self.logging.hedra.sync.debug(f'{self._graph_metadata_log_string} - Generating - {stages_count} - stages')
        self.logging.filesystem.sync['hedra.core'].debug(f'{self._graph_metadata_log_string} - Generating - {stages_count} - stages')

        self.instances_by_type = {}

        for stage in stages.values():
            self.instances_by_type[stage.stage_type] = []

        stage_types_count = len(self.instances_by_type)
        self.logging.hedra.sync.debug(f'{self._graph_metadata_log_string} - Found - {stage_types_count} - unique stage types')
        self.logging.filesystem.sync['hedra.core'].debug(f'{self._graph_metadata_log_string} - Found - {stage_types_count} - unique stage types')

        self.generated_stages: Dict[str, Stage] = {
            stage_name: stage() for stage_name, stage in stages.items() 
        }

        generated_hooks = {}
        for stage in self.generated_stages.values():

            if stage.name in self.graph_skipped_stages:
                stage.skip = True

            stage.core_config = self.core_config
            stage.graph_name = self.graph_name
            stage.graph_path = self.graph_path
            stage.graph_id = self.graph_id

            stage.workers = self.cpus
            stage.worker_id = self.worker_id

            for hook_shortname, hook in registrar.reserved[stage.name].items():
                hook._call = hook._call.__get__(stage, stage.__class__)
                setattr(stage, hook_shortname, hook._call)

            stage = set_stage_hooks(stage, generated_hooks)

            for hook_type in stage.hooks:
                for hook in stage.hooks[hook_type]: 
                    self.hooks_by_type[hook.hook_type][hook.name] = hook
                
            self.instances_by_type[stage.stage_type].append(stage)

        events_graph = EventGraph(self.hooks_by_type)
        events_graph.hooks_to_events().assemble_graph().apply_graph_to_events()

        self.logging.hedra.sync.debug(f'{self._graph_metadata_log_string} - Successfully generated - {stages_count} - stages')
        self.logging.filesystem.sync['hedra.core'].debug(f'{self._graph_metadata_log_string} - Successfully generated - {stages_count} - stages')

    def build_transitions_graph(self, topological_generations: List[List[str]], graph: networkx.DiGraph) -> List[TransitionGroup]:

        self.logging.hedra.sync.debug(f'Buiding transitions matrix')
        self.logging.filesystem.sync['hedra.core'].debug(f'Buiding transitions matrix')

        transitions: List[TransitionGroup] = []
        plugins: Dict[PluginType, Dict[str, Union[EnginePlugin, ReporterPlugin]]] = {
            PluginType.ENGINE: {},
            PluginType.OPTIMIZER: {},
            PluginType.PERSONA: {},
            PluginType.REPORTER: {}
        }

        for isolate_stage_name in networkx.isolates(graph):
            raise IsolatedStageError(
                self.generated_stages.get(isolate_stage_name)
            )
        
        for generation in topological_generations:

            generation_transitions = TransitionGroup()
            generation_transitions.cpu_pool_size = self.cpus

            stage_pool_size = self.cpus

            stages = {
                stage_name: self.generated_stages.get(stage_name) for stage_name in generation
            }
            parallel_stages = []

            no_workers_stages = [
                StageTypes.IDLE
            ]

            stages_count = len(stages)
            self.logging.hedra.sync.debug(f'{self._graph_metadata_log_string} - Provisioning workers - {stages_count} - stages')
            self.logging.filesystem.sync['hedra.core'].debug(f'{self._graph_metadata_log_string} - Provisioning workers- {stages_count} - stages')

            for stage in stages.values():

                stage.total_pool_cpus = self.cpus

                for plugin_name, plugin in stage.plugins.items():
                    plugins[plugin.type][plugin_name] = plugin

                if stage.allow_parallel is False and stage.stage_type not in no_workers_stages:
                    stage.workers = 1
                    stage_pool_size -= 1

                    self.logging.hedra.sync.debug(f'{self._graph_metadata_log_string} - Stage - {stage.name} - provisioned - {stage.workers} - workers')
                    self.logging.filesystem.sync['hedra.core'].debug(f'{self._graph_metadata_log_string} - Stage - {stage.name} - provisioned - {stage.workers} - workers')

                    stage.executor = BatchExecutor(stage.workers)
                    self.executors.append(stage.executor)

                else:
                    parallel_stages.append((
                        stage.name,
                        stage
                    ))
            
            if len(parallel_stages) > 0:
                batch_executor = BatchExecutor(max_workers=stage_pool_size)
                self.executors.append(batch_executor)

                batched_stages: List[Tuple[str, Stage, int]] = batch_executor.partion_stage_batches(parallel_stages)
                
                for _, stage, assigned_workers_count in batched_stages:

                    self.logging.hedra.sync.debug(f'{self._graph_metadata_log_string} - Stage - {stage.name} - provisioned - {assigned_workers_count} - workers')
                    self.logging.filesystem.sync['hedra.core'].debug(f'{self._graph_metadata_log_string} - Stage - {stage.name} - provisioned - {assigned_workers_count} - workers')

                    stage.workers = assigned_workers_count
                    stage.executor = BatchExecutor(max_workers=assigned_workers_count)
                    self.executors.append(stage.executor)

                    stages[stage.name] = stage

                batch_executor.close()

            for stage in stages.values():

                stage.plugins_by_type = plugins
                stage.dispatcher.assemble_action_and_task_subgraphs()

                neighbors = list(graph.neighbors(stage.name))

                neighbors_count = len(neighbors)
                self.logging.hedra.sync.debug(f'{self._graph_metadata_log_string} - Discovered - {neighbors_count} - neighboring stages for stage - {stage.name}')
                self.logging.filesystem.sync['hedra.core'].debug(f'{self._graph_metadata_log_string} - Discovered - {neighbors_count} - neighboring stages for stage - {stage.name}')
                
                for neighbor in neighbors:
                    neighbor_stage = self.generated_stages.get(neighbor)

                    transition_action: TransitionMetadata = self.transition_types.get((
                        stage.stage_type,
                        neighbor_stage.stage_type
                    ))

                    self.logging.hedra.sync.debug(f'{self._graph_metadata_log_string} - Created transition from - {stage.name} - to - {neighbor_stage.name}')
                    self.logging.filesystem.sync['hedra.core'].debug(f'{self._graph_metadata_log_string} - Created transition from - {stage.name} - to - {neighbor_stage.name}')

                    transition = Transition(
                        transition_action,
                        stage,
                        neighbor_stage
                    )

                    if stage.stage_type == StageTypes.EXECUTE:
                        self.execute_stages.append(
                            stage
                        )

                    if stage.stage_type == StageTypes.SUBMIT and stage.stream:
                        self.streaming_submit_stages.append(
                            stage
                        )


                    if transition_action.is_valid is False:
                        raise InvalidTransitionError(
                            transition.from_stage,
                            transition.to_stage
                        )

                    transition.predecessors = list(graph.predecessors(stage.name))
                    transition.descendants = list(graph.successors(stage.name))
                    
                    self.adjacency_list[stage.name].append(transition)

                    self.edges_by_name[(transition.from_stage.name, transition.to_stage.name)] = transition.edge

                    generation_transitions.add_transition(transition)

            if generation_transitions.count > 0:
                transitions.append(generation_transitions)

        for transition_group in transitions:
            transition_group.adjacency_list = self.adjacency_list
            transition_group.edges_by_name = self.edges_by_name

            for transition in transition_group:
                transition.adjacency_list = self.adjacency_list
                transition.edges_by_name = self.edges_by_name

        self.logging.hedra.sync.debug(f'{self._graph_metadata_log_string} - Transition matrix assemmbly complete')
        self.logging.filesystem.sync['hedra.core'].debug(f'{self._graph_metadata_log_string} - Transition matrix assemmbly complete')

        return transitions

    def map_to_setup_stages(self, graph: networkx.DiGraph) -> None:

        
        self.logging.hedra.sync.debug(f'{self._graph_metadata_log_string} - Mapping stages to requisite Setup stages')
        self.logging.filesystem.sync['hedra.core'].debug(f'{self._graph_metadata_log_string} - Mapping stages to requisite Setup stages')

        idle_stages = self.instances_by_type.get(StageTypes.IDLE)
        for idle_stage in idle_stages:
            idle_stage.context = SimpleContext()
            idle_stage.context.stages = {}
            idle_stage.context.visited = []
            idle_stage.context.results = {}
            idle_stage.context.results_stages = []
            idle_stage.context.summaries = {}
            idle_stage.context.paths = {}
            idle_stage.context.path_lengths = {}
            
            idle_stage.name = idle_stage.__class__.__name__


        complete_stage = self.instances_by_type.get(StageTypes.COMPLETE)[0]

        stages_by_type = defaultdict(dict)
        for stage_type in self.instances_by_type:
            for stage in self.instances_by_type[stage_type]:
                stages_by_type[stage_type][stage.name] = stage

        self.all_paths: Dict[str, List[str]] = {}
        
        for execute_stage in self.execute_stages:

            if execute_stage.context is None:
                execute_stage.context = SimpleContext()

            for streaming_submit_stage in self.streaming_submit_stages:

                has_path = networkx.has_path(
                    graph, 
                    execute_stage.name, 
                    streaming_submit_stage.name
                )
                
                if has_path:
                    if execute_stage.context['execute_stage_stream_configs'] is None:
                        execute_stage.context['execute_stage_stream_configs'] = [
                            streaming_submit_stage.config
                        ]

                    else:
                        execute_stage.context['execute_stage_stream_configs'].append(
                            streaming_submit_stage.config
                        )


        for stage_type in StageTypes:

            idle_stage.context.stages[stage_type] = {}

            for stage in self.instances_by_type.get(stage_type, []):

                stage_name = stage.__class__.__name__

                for neighbor in self.adjacency_list[stage.name]:
                    self.edges_by_name[(stage.name, neighbor.edge.destination.name)].stages_by_type = stages_by_type
                    paths = networkx.all_simple_paths(graph, stage_name, complete_stage.name)
                
                    stage_paths = []
                    for path in paths:
                        stage_paths.extend(path)
                    
                    self.all_paths[stage_name] = stage_paths
                    
                    path_lengths = networkx.all_pairs_shortest_path_length(graph)

                    stage_path_lengths = {}
                    for path_stage_name, path_lengths_set in path_lengths:

                        del path_lengths_set[path_stage_name]
                        stage_path_lengths[path_stage_name] = path_lengths_set

                    self.edges_by_name[(stage.name, neighbor.edge.destination.name)].path_lengths[stage_name] = stage_path_lengths.get(stage_name)

        for stage in self.generated_stages.values():
            for neighbor in self.adjacency_list[stage.name]:
                self.edges_by_name[(stage.name, neighbor.edge.destination.name)].all_paths = self.all_paths

        self.logging.hedra.sync.debug(f'{self._graph_metadata_log_string} - Mapped stages to requisite Setup stages')
        self.logging.filesystem.sync['hedra.core'].debug(f'{self._graph_metadata_log_string} - Mapped stages to requisite Setup stages')

    def apply_config_to_load_hooks(
        self,
        graph: networkx.DiGraph
    ):

        setup_stages = self.instances_by_type.get(StageTypes.SETUP)

        for stage in setup_stages:
            for load_hook in stage.hooks[HookType.LOAD]:
                load_hook: LoadHook = load_hook

                load_hook.parser_config = stage.config

        non_setup_stages: List[Stage] = []
        for stage in self.generated_stages.values():
            if stage not in setup_stages:
                non_setup_stages.append(stage)

        path_lengths = dict(networkx.all_pairs_shortest_path_length(graph))

        for stage in non_setup_stages:

            setup_to_stage_path_lengths: Dict[str, int] = {}

            for setup_stage in setup_stages:
                if networkx.has_path(graph, setup_stage.name, stage.name):
                    path_length = path_lengths[setup_stage.name][stage.name]
                    setup_to_stage_path_lengths[setup_stage.name] = path_length

            if len(setup_to_stage_path_lengths):
                minimum_distance_setup_stage = min(
                    setup_to_stage_path_lengths,
                    key=lambda stage_name: setup_to_stage_path_lengths.get(stage_name)
                )

                setup_stage = self.generated_stages.get(minimum_distance_setup_stage)

                for load_hook in stage.hooks[HookType.LOAD]:
                    load_hook.parser_config = setup_stage.config

    
    def create_error_transition(
        self, 
        source_stage: Stage,
        error: Exception
    ) -> Transition:
            
        error_transition = self.transition_types.get((
            source_stage.stage_type,
            StageTypes.ERROR
        ))

        error_stage = Error()
        error_stage.graph_name = self.graph_name
        error_stage.graph_id = self.graph_id
        error_stage.error = error

        return Transition(
            error_transition,
            error_stage,
            error_stage
        )


from __future__ import annotations
import asyncio
import inspect
from typing import (
    Any, 
    Awaitable, 
    Generic, 
    TypeVar
)
from hedra.core.engines.client.config import Config
from hedra.core.engines.types.custom.client import MercuryCustomClient as CustomSession
from hedra.core.engines.types.common import Timeouts
from hedra.core.engines.client.store import ActionsStore
from hedra.plugins.types.plugin_types import PluginType
from hedra.plugins.types.common.event import Event
from hedra.plugins.types.common.plugin_hook import PluginHook
from hedra.plugins.types.common.types import PluginHooks
from hedra.plugins.types.common.plugin import Plugin
from hedra.plugins.types.common.registrar import plugin_registrar
from .action import Action
from .result import Result


A = TypeVar('A')
R = TypeVar('R')


class EnginePlugin(Generic[A, R], Plugin):
    action: Action[A] = None
    result: Result[R] = None
    event: Event[Result[R]] = Event
    security_context: Any = None
    initialized: bool = False
    type=PluginType.ENGINE
    name: str=None

    def __init__(self, config: Config) -> None:

        super(
            EnginePlugin,
            self
        ).__init__()

        self.hooks = {}
        self.request_type = self.__class__.__name__
        self.next_name = None
        self.intercept = False
        self.waiter = None
        self.actions: ActionsStore = None
        self.registered = {}
        self.metadata_string: str = None
        self.name = self.name

        self.config = config
        

        self.action_type = A
        self.request_type = R

        methods = inspect.getmembers(self, predicate=inspect.ismethod) 
        for _, method in methods:

                method_name = method.__qualname__
                hook: PluginHook = plugin_registrar.all.get(method_name)
                
                if hook:
                    hook.call = hook.call.__get__(self, self.__class__)
                    setattr(self, hook.shortname, hook.call)

                    self.hooks[hook.hook_type] = hook

 
        self.session = CustomSession[A, R](
            self,
            concurrency=config.batch_size,
            timeouts=Timeouts(
                connect_timeout=config.connect_timeout,
                total_timeout=config.request_timeout
            ),
            reset_connections=config.reset_connections
        )

    def __getattr__(self, attribute_name: str):

        session = object.__getattribute__(self, 'session')

        if hasattr(session, attribute_name):
            return getattr(session, attribute_name)
            
        else:
            # Default behaviour
            return object.__getattribute__(self, attribute_name)

    async def execute(self, *args, **kwargs) -> Awaitable[Result[R]]:
        
        if self.registered.get(self.next_name) is None:
            action: Action[A] = self.action(
                self.next_name,
                *args,
                **kwargs
            )

            action.plugin_type = self.name
            
            await self.session.prepare(action)

            if self.intercept:
                self.actions.store(self.next_name, action, self)
                
                loop = asyncio.get_event_loop()
                self.waiter = loop.create_future()
                await self.waiter

        return await self.session.execute_prepared_request(
            self.session.registered.get(self.next_name)
        )

    async def close(self):
        close_hook = self.hooks.get(PluginHooks.ON_ENGINE_CLOSE)
        await close_hook.call()


    



import uuid
from typing import List, Dict
from hedra.logging import HedraLogger
from hedra.reporting.experiment.experiments_collection import ExperimentMetricsCollectionSet
from hedra.reporting.processed_result.types.base_processed_result import BaseProcessedResult
from hedra.reporting.metric.stage_streams_set import StageStreamsSet
from hedra.reporting.metric import MetricsSet
from hedra.reporting.system.system_metrics_set import (
    SystemMetricsSet,
    SessionMetricsCollection,
    SystemMetricsCollection
)
from .cosmosdb_config import CosmosDBConfig


try:
    from azure.cosmos.aio import CosmosClient
    from azure.cosmos import PartitionKey
    has_connector = True
except Exception:
    CosmosClient = None
    PartitionKey = None
    has_connector = False


class CosmosDB:

    def __init__(self, config: CosmosDBConfig) -> None:
        self.account_uri = config.account_uri
        self.account_key = config.account_key

        self.database_name = config.database
        self.events_container_name = config.events_container
        self.metrics_container_name = config.metrics_container
        self.streams_container_name = config.streams_container

        self.experiments_container_name = config.experiments_container
        self.variants_container_name = f'{config.experiments_container}_variants'
        self.mutations_container_name = f'{config.experiments_container}_mutations'

        self.shared_metrics_container_name = f'{self.metrics_container_name}_metrics'
        self.custom_metrics_container_name = f'{self.metrics_container_name}_custom'
        self.errors_container_name = f'{self.metrics_container}_errors'

        self.session_system_metrics_container_name = f'{config.system_metrics_container}_session'
        self.stage_system_metrics_container_name = f'{config.system_metrics_container}_stage'

        self.events_partition_key = config.events_partition_key
        self.metrics_partition_key = config.metrics_partition_key
        self.streams_partition_key = config.streams_partition_key
        self.system_metrics_partition = config.system_metrics_partition

        self.experiments_partition_key = config.experiments_partition_key
        self.variants_partition_key = config.variants_partition_key
        self.mutations_partition_key = config.mutations_partition_key


        self.analytics_ttl = config.analytics_ttl

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.logger = HedraLogger()
        self.logger.initialize()

        self.events_container = None
        self.metrics_container = None
        self.streams_container = None

        self.session_system_metrics_container = None
        self.stage_system_metrics_container = None

        self.experiments_container = None
        self.variants_container = None
        self.mutations_container = None

        self.shared_metrics_container = None
        self.custom_metrics_container = None
        self.errors_container = None
        
        self.client = None
        self.database = None

    async def connect(self):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Connecting to CosmosDB')

        self.client = CosmosClient(
            self.account_uri,
            credential=self.account_key
        )


        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Connected to CosmosDB')
        
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Database - {self.database_name} - if not exists')
        self.database = await self.client.create_database_if_not_exists(self.database_name)

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created or set Database - {self.database_name}')

    async def submit_session_system_metrics(self, system_metrics_sets: List[SystemMetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Session System Metrics container - {self.session_system_metrics_container_name} with Partition Key /{self.system_metrics_partition} if not exists')
        self.session_system_metrics_container = await self.database.create_container_if_not_exists(
            self.session_system_metrics_container_name,
            PartitionKey(f'/{self.system_metrics_partition}')
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created or set Session System Metrics container - {self.session_system_metrics_container_name} with Partition Key /{self.system_metrics_partition}')

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Session System Metrics to container - {self.session_system_metrics_container_name} with Partition Key /{self.system_metrics_partition}')
        
        rows: List[SessionMetricsCollection] = []

        for metrics_set in system_metrics_sets:

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Session System Metrics Set - {metrics_set.system_metrics_set_id}')

            for monitor_metrics in metrics_set.session_cpu_metrics.values():
                rows.append(monitor_metrics)
                
            for  monitor_metrics in metrics_set.session_memory_metrics.values():
                rows.append(monitor_metrics)

        for metrics_set in rows:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Session System Metrics Group - {metrics_set.name}:{metrics_set.group}')

            await self.session_system_metrics_container.upsert_item({
                'id': str(uuid.uuid4()),
                **metrics_set.record
            })
            
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Session System Metrics to container - {self.session_system_metrics_container_name} with Partition Key /{self.system_metrics_partition}')

    async def submit_stage_system_metrics(self, system_metrics_sets: List[SystemMetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Stage System Metrics container - {self.stage_system_metrics_container_name} with Partition Key /{self.system_metrics_partition} if not exists')
        self.stage_system_metrics_container = await self.database.create_container_if_not_exists(
            self.stage_system_metrics_container_name,
            PartitionKey(f'/{self.system_metrics_partition}')
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created or set Stage System Metrics container - {self.stage_system_metrics_container_name} with Partition Key /{self.system_metrics_partition}')

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Stage System Metrics to container - {self.stage_system_metrics_container_name} with Partition Key /{self.system_metrics_partition}')
        
        rows: List[SystemMetricsCollection] = []

        for metrics_set in system_metrics_sets:

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Stage System Metrics Set - {metrics_set.system_metrics_set_id}')

            cpu_metrics = metrics_set.cpu
            memory_metrics = metrics_set.memory

            for stage_name, stage_cpu_metrics in  cpu_metrics.metrics.items():

                for monitor_metrics in stage_cpu_metrics.values():
                    rows.append(monitor_metrics.record)

                stage_memory_metrics = memory_metrics.metrics.get(stage_name)
                for monitor_metrics in stage_memory_metrics.values():
                    rows.append(monitor_metrics.record)

                stage_mb_per_vu_metrics = metrics_set.mb_per_vu.get(stage_name)
                
                if stage_mb_per_vu_metrics:
                    rows.append(stage_mb_per_vu_metrics.record)

        for metrics_set in rows:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Stage System Metrics Group - {metrics_set.name}:{metrics_set.group}')

            await self.stage_system_metrics_container.upsert_item({
                'id': str(uuid.uuid4()),
                **metrics_set.record
            })
            
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Stage System Metrics to container - {self.stage_system_metrics_container_name} with Partition Key /{self.system_metrics_partition}')

    async def submit_streams(self, stream_metrics: Dict[str, StageStreamsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Streams container - {self.streams_container_name} with Partition Key /{self.streams_partition_key} if not exists')
        self.streams_container = await self.database.create_container_if_not_exists(
            self.streams_container_name,
            PartitionKey(f'/{self.streams_partition_key}')
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created or set Streams container - {self.streams_container_name} with Partition Key /{self.streams_partition_key}')

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Streams to container - {self.streams_container_name} with Partition Key /{self.streams_partition_key}')
        for stage_name, stream in stream_metrics.items():
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Streams Set - {stage_name}:{stream.stream_set_id}')

            for group_name, group in stream.grouped.items():
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Streams Group - {group_name}:{stream.stream_set_id}')

                await self.streams_container.upsert_item({
                    'id': str(uuid.uuid4()),
                    'name': f'{stage_name}_stream',
                    'stage': stage_name,
                    'group': group_name,
                    **group
                })
            
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Streams to container - {self.streams_container_name} with Partition Key /{self.streams_partition_key}')

    async def submit_experiments(self, experiment_metrics: ExperimentMetricsCollectionSet):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Experiments container - {self.experiments_container_name} with Partition Key /{self.experiments_partition_key} if not exists')
        self.experiments_container = await self.database.create_container_if_not_exists(
            self.experiments_container_name,
            PartitionKey(f'/{self.experiments_partition_key}')
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created or set Experiment container - {self.experiments_container_name} with Partition Key /{self.experiments_partition_key}')

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Experiment to container - {self.experiments_container_name} with Partition Key /{self.experiments_partition_key}')
        for experiment in experiment_metrics.experiments:
            await self.experiments_container.upsert_item({
                'id': str(uuid.uuid4()),
                **experiment
            })
        
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Experiments to container - {self.experiments_container_name} with Partition Key /{self.experiments_partition_key}')

    async def submit_variants(self, experiment_metrics: ExperimentMetricsCollectionSet):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Variants container - {self.variants_container_name} with Partition Key /{self.variants_partition_key} if not exists')
        self.variants_container = await self.database.create_container_if_not_exists(
            self.variants_container_name,
            PartitionKey(f'/{self.variants_partition_key}')
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created or set Variants container - {self.variants_container_name} with Partition Key /{self.variants_partition_key}')

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Variant to container - {self.variants_container_name} with Partition Key /{self.variants_partition_key}')
        for variant in experiment_metrics.variants:
            await self.variants_container.upsert_item({
                'id': str(uuid.uuid4()),
                **variant
            })
        
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Variants to container - {self.variants_container_name} with Partition Key /{self.variants_partition_key}')

    async def submit_mutations(self, experiment_metrics: ExperimentMetricsCollectionSet):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Mutations container - {self.mutations_container_name} with Partition Key /{self.mutations_partition_key} if not exists')
        self.mutations_container = await self.database.create_container_if_not_exists(
            self.mutations_container_name,
            PartitionKey(f'/{self.mutations_partition_key}')
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created or set Mutations container - {self.mutations_container_name} with Partition Key /{self.mutations_partition_key}')

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Mutation to container - {self.mutations_container_name} with Partition Key /{self.mutations_partition_key}')
        for mutation in experiment_metrics.mutations:
            await self.mutations_container.upsert_item({
                'id': str(uuid.uuid4()),
                **mutation
            })
        
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Mutations to container - {self.mutations_container_name} with Partition Key /{self.mutations_partition_key}')


    async def submit_events(self, events: List[BaseProcessedResult]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Events container - {self.events_container_name} with Partition Key /{self.events_partition_key} if not exists')
        self.events_container = await self.database.create_container_if_not_exists(
            self.events_container_name,
            PartitionKey(f'/{self.events_partition_key}')
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created or set Events container - {self.events_container_name} with Partition Key /{self.events_partition_key}')

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Events to container - {self.events_container_name} with Partition Key /{self.events_partition_key}')
        for event in events:
            await self.events_container.upsert_item({
                'id': str(uuid.uuid4()),
                **event.record
            })

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Events to container - {self.events_container_name} with Partition Key /{self.events_partition_key}')

    async def submit_common(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Shared Metrics container - {self.shared_metrics_container_name} with Partition Key /{self.metrics_partition_key} if not exists')
        self.shared_metrics_container = await self.database.create_container_if_not_exists(
            self.shared_metrics_container_name,
            PartitionKey(f'/{self.metrics_partition_key}')
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created or set Shared Metrics container - {self.shared_metrics_container_name} with Partition Key /{self.metrics_partition_key}')


        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Shared Metrics to container - {self.shared_metrics_container_name} with Partition Key /{self.metrics_partition_key}')
        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            await self.shared_metrics_container.upsert_item({
                'id': str(uuid.uuid4()),
                'name': metrics_set.name,
                'stage': metrics_set.stage,
                'group': 'common',
                **metrics_set.common_stats
            })

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Shared Metrics to container - {self.shared_metrics_container_name} with Partition Key /{self.metrics_partition_key}')
        
    async def submit_metrics(self, metrics: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Metrics container - {self.metrics_container_name} with Partition Key /{self.metrics_partition_key} if not exists')
        self.metrics_container = await self.database.create_container_if_not_exists(
            self.metrics_container_name,
            PartitionKey(f'/{self.metrics_partition_key}')
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created or set Metrics container - {self.metrics_container_name} with Partition Key /{self.metrics_partition_key}')

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Metrics to container - {self.metrics_container_name} with Partition Key /{self.metrics_partition_key}')
        for metrics_set in metrics:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            for group_name, group in metrics_set.groups.items():
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metrics Group - {group_name}:{group.metrics_group_id}')

                await self.metrics_container.upsert_item({
                    'id': str(uuid.uuid4()),
                    'group': group_name,
                    **group.record
                })
            
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Metrics to container - {self.metrics_container_name} with Partition Key /{self.metrics_partition_key}')

    async def submit_custom(self, metrics_sets: List[MetricsSet]):
        
        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Custom Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Custom Metrics Group - Xuarom')

            custom_metrics_container_name = f'{self.custom_metrics_container_name}_metrics'
            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Custom Metrics container - {custom_metrics_container_name} with Partition Key /{self.metrics_partition_key} if not exists')

            custom_container = await self.database.create_container_if_not_exists(
                custom_metrics_container_name,
                PartitionKey(f'/{self.metrics_partition_key}')
            )

            self.custom_metrics_container = custom_container


            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created or set Custom Metrics container - {custom_metrics_container_name} with Partition Key /{self.metrics_partition_key}')

            await custom_container.upsert_item({
                'id': str(uuid.uuid4()),
                'name': metrics_set.name,
                'stage': metrics_set.stage,
                'group': 'custom',
                **{
                    custom_metric_name: custom_metric.metric_value for custom_metric_name, custom_metric in metrics_set.custom_metrics.items()
                }
            })

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Metrics to container - {self.metrics_container_name} with Partition Key /{self.metrics_partition_key}')


    async def submit_errors(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Error Metrics container - {self.errors_container_name} with Partition Key /{self.metrics_partition_key} if not exists')
        self.errors_container = await self.database.create_container_if_not_exists(
            self.errors_container_name,
            PartitionKey(f'/{self.metrics_partition_key}')
        )
        
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created or set Error Metrics container - {self.errors_container_name} with Partition Key /{self.metrics_partition_key}')

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Error Metrics to container - {self.errors_container_name} with Partition Key /{self.metrics_partition_key}')
        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Error Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            for error in metrics_set.errors:
                await self.metrics_container.upsert_item({
                    'id': str(uuid.uuid4()),
                    'name': metrics_set.name,
                    'stage': metrics_set.stage,
                    'error_message': error.get('message'),
                    'error_count': error.get('count')
                })

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Error Metrics to container - {self.errors_container_name} with Partition Key /{self.metrics_partition_key}')

    async def close(self):

        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Closing session - {self.session_uuid}')
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Closing connection to CosmosDB')

        await self.client.close()
        
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Closed connection to CosmosDB')


import tdigest


class MedianAbsoluteDeviation:

    __slots__ = (
        'approx_median',
        'update_interval',
        'records_processed',
        'values',
        'deviations',
        'constant'
    )
    
    def __init__(self, update_interval=5, sensitivity=0.01, compression_factor=25):
        self.approx_median = None
        self.update_interval = update_interval
        self.records_processed = 0
        self.values = tdigest.TDigest()
        self.deviations = tdigest.TDigest(delta=sensitivity,K=compression_factor)
        self.constant = 1.4826

    def update(self, new_value):
        if self.approx_median is None:
            self.approx_median = new_value

        self.values.update(new_value)
        deviation = abs(self.approx_median - new_value)
        self.deviations.update(deviation)

        self.records_processed += 1
        
        if self.records_processed%self.update_interval == 0:
            self.approx_median = self.values.percentile(50)

    def get(self):
        if len(self.deviations) > 0:
            return self.deviations.percentile(50) * self.constant
        else:
            return 0
